"""CFNgin deploy action."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

from typing_extensions import Literal

from ..exceptions import (
    CancelExecution,
    CfnginBucketRequired,
    MissingParameterException,
    StackDidNotChange,
    StackDoesNotExist,
)
from ..hooks.utils import handle_hooks as _handle_hooks
from ..plan import Graph, Plan, Step
from ..providers.base import Template
from ..status import (
    INTERRUPTED,
    PENDING,
    SUBMITTED,
    WAITING,
    CompleteStatus,
    DidNotChangeStatus,
    DoesNotExistInCloudFormation,
    FailedStatus,
    NotSubmittedStatus,
    NotUpdatedStatus,
    SkippedStatus,
    SubmittedStatus,
)
from .base import STACK_POLL_TIME, BaseAction, build_walker

if TYPE_CHECKING:
    from mypy_boto3_cloudformation.type_defs import ParameterTypeDef, StackTypeDef

    from ...config.models.cfngin import CfnginHookDefinitionModel
    from ...context import CfnginContext
    from ...core.providers.aws.type_defs import TagTypeDef
    from ..blueprints.base import Blueprint
    from ..providers.aws.default import Provider
    from ..stack import Stack
    from ..status import Status

LOGGER = logging.getLogger(__name__)

DESTROYED_STATUS = CompleteStatus("stack destroyed")
DESTROYING_STATUS = SubmittedStatus("submitted for destruction")


def build_stack_tags(stack: Stack) -> List[TagTypeDef]:
    """Build a common set of tags to attach to a stack."""
    return [{"Key": t[0], "Value": t[1]} for t in stack.tags.items()]


def should_update(stack: Stack) -> bool:
    """Test whether a stack should be submitted for updates to CloudFormation.

    Args:
        stack: The stack object to check.

    """
    if stack.locked:
        if not stack.force:
            LOGGER.debug(
                "%s:locked and not in --force list; refusing to update", stack.name
            )
            return False
        LOGGER.debug("%s:locked but is in --force list", stack.name)
    return True


def should_submit(stack: Stack) -> bool:
    """Test whether a stack should be submitted to CF for update/create.

    Args:
        stack: The stack object to check.

    """
    if stack.enabled:
        return True

    LOGGER.debug("%s:skipped; stack is not enabled", stack.name)
    return False


def should_ensure_cfn_bucket(outline: bool, dump: bool) -> bool:
    """Test whether access to the cloudformation template bucket is required.

    Args:
        outline: The outline action.
        dump: The dump action.

    Returns:
        If access to CF bucket is needed, return True.

    """
    return not outline and not dump


def _resolve_parameters(
    parameters: Dict[str, Any], blueprint: Blueprint
) -> Dict[str, Any]:
    """Resolve CloudFormation Parameters for a given blueprint.

    Given a list of parameters, handles:
        - discard any parameters that the blueprint does not use
        - discard any empty values
        - convert booleans to strings suitable for CloudFormation

    Args:
        parameters: A dictionary of parameters provided by the stack definition.
        blueprint: A Blueprint object that is having the parameters applied to it.

    Returns:
        The resolved parameters.

    """
    params: Dict[str, Any] = {}
    for key, value in parameters.items():
        if key not in blueprint.parameter_definitions:
            LOGGER.debug("blueprint %s does not use parameter %s", blueprint.name, key)
            continue
        if value is None:
            LOGGER.debug(
                "got NoneType value for parameter %s; not submitting it "
                "to cloudformation, default value should be used",
                key,
            )
            continue
        if isinstance(value, bool):
            LOGGER.debug('converting parameter %s boolean "%s" to string', key, value)
            value = str(value).lower()
        params[key] = value
    return params


class UsePreviousParameterValue:
    """Class used to indicate a Parameter should use it's existing value."""


def _handle_missing_parameters(
    parameter_values: Dict[str, Any],
    all_params: List[str],
    required_params: List[str],
    existing_stack: Optional[StackTypeDef] = None,
) -> List[Tuple[str, Any]]:
    """Handle any missing parameters.

    If an existing_stack is provided, look up missing parameters there.

    Args:
        parameter_values: key/value dictionary of stack definition parameters.
        all_params: A list of all the parameters used by the template/blueprint.
        required_params: A list of all the parameters required by the template/blueprint.
        existing_stack: A dict representation of the stack. If provided, will be
            searched for any missing parameters.

    Returns:
        The final list of key/value pairs returned as a list of tuples.

    Raises:
        MissingParameterException: Raised if a required parameter is
            still missing.

    """
    missing_params = list(set(all_params) - set(parameter_values.keys()))
    if existing_stack and "Parameters" in existing_stack:
        stack_parameters = [
            param["ParameterKey"]
            for param in existing_stack["Parameters"]
            if "ParameterKey" in param
        ]
        for param in missing_params:
            if param in stack_parameters:
                LOGGER.debug(
                    "using previous value for parameter %s from existing stack", param
                )
                parameter_values[param] = UsePreviousParameterValue
    final_missing = list(set(required_params) - set(parameter_values.keys()))
    if final_missing:
        raise MissingParameterException(final_missing)

    return list(parameter_values.items())


def handle_hooks(
    stage: Literal["post_deploy", "pre_deploy"],
    hooks: List[CfnginHookDefinitionModel],
    provider: Provider,
    context: CfnginContext,
    *,
    dump: Union[bool, str] = False,
    outline: bool = False,
) -> None:
    """Handle pre/post hooks.

    Args:
        stage: The name of the hook stage - pre_deploy/post_deploy.
        hooks: A list of dictionaries containing the hooks to execute.
        provider: The provider the current stack is using.
        context: The current CFNgin context.
        dump: Whether running with dump set or not.
        outline: Whether running with outline set or not.

    """
    if not outline and not dump and hooks:
        _handle_hooks(stage=stage, hooks=hooks, provider=provider, context=context)


class Action(BaseAction):
    """Responsible for building & deploying CloudFormation stacks.

    Generates the deploy plan based on stack dependencies (these dependencies
    are determined automatically based on output lookups from other stacks).

    The plan can then either be printed out as an outline or executed. If
    executed, each stack will get launched in order which entails:

    - Pushing the generated CloudFormation template to S3 if it has changed
    - Submitting either a create or update of the given stack to the
      :class:`runway.cfngin.providers.base.BaseProvider`.

    Attributes:
        upload_explicitly_disabled: Explicitly disable uploading rendered templates
            to S3.

    """

    DESCRIPTION = "Create/Update stacks"
    NAME = "deploy"

    upload_explicitly_disabled: bool = False

    @property
    def upload_disabled(self) -> bool:
        """Whether the CloudFormation template should be uploaded to S3."""
        if self.upload_explicitly_disabled:
            return True
        if not self.bucket_name:
            return True
        return False

    @upload_disabled.setter
    def upload_disabled(self, value: bool) -> None:
        """Set the value of upload_disabled.

        Raises:
            CfnginBucketRequired: Attempted to explicitly enable upload but cfngin_bucket
                not defined.

        """
        if not value and not self.bucket_name:
            raise CfnginBucketRequired(
                config_path=self.context.config_path,
                reason="upload_disabled explicitly set to False",
            )
        self.upload_explicitly_disabled = value

    @staticmethod
    def build_parameters(
        stack: Stack, provider_stack: Optional[StackTypeDef] = None
    ) -> List[ParameterTypeDef]:
        """Build the CloudFormation Parameters for our stack.

        Args:
            stack: A CFNgin stack.
            provider_stack: An optional CFNgin provider object.

        Returns:
            The parameters for the given stack

        """
        resolved = _resolve_parameters(stack.parameter_values, stack.blueprint)
        required_parameters = list(stack.required_parameter_definitions)
        all_parameters = list(stack.all_parameter_definitions)
        parameters = _handle_missing_parameters(
            resolved, all_parameters, required_parameters, provider_stack
        )

        param_list: List[ParameterTypeDef] = []

        for key, value in parameters:
            param_dict: ParameterTypeDef = {"ParameterKey": key}
            if value is UsePreviousParameterValue:
                param_dict["UsePreviousValue"] = True
            else:
                param_dict["ParameterValue"] = str(value)

            param_list.append(param_dict)

        return param_list

    def _destroy_stack(
        self, stack: Stack, *, status: Optional[Status] = None, **_: Any
    ) -> Status:
        """Delete a CloudFormation stack.

        Used to remove stacks that exist in the persistent graph but not
        have been removed from the "local" graph.

        Args:
            stack: Stack to be deleted.
            status: The Stack's status represented by a CFNgin status object.

        """
        wait_time = 0 if status is PENDING else STACK_POLL_TIME
        if self.cancel.wait(wait_time):
            return INTERRUPTED

        provider = self.build_provider()

        try:
            stack_data = provider.get_stack(stack.fqn)
        except StackDoesNotExist:
            LOGGER.debug("%s:stack does not exist", stack.fqn)
            if status == SUBMITTED:
                return DESTROYED_STATUS
            return DoesNotExistInCloudFormation()

        LOGGER.debug(
            "%s:provider status: %s",
            provider.get_stack_name(stack_data),
            provider.get_stack_status(stack_data),
        )
        try:
            if provider.is_stack_being_destroyed(stack_data):
                return DESTROYING_STATUS
            if provider.is_stack_destroyed(stack_data):
                return DESTROYED_STATUS
            wait = stack.in_progress_behavior == "wait"
            if wait and provider.is_stack_in_progress(stack_data):
                return WAITING
            if provider.is_stack_destroy_possible(stack_data):
                LOGGER.debug("%s:destroying stack", stack.fqn)
                provider.destroy_stack(stack_data, action="deploy")
                return DESTROYING_STATUS
            LOGGER.critical(
                "%s: %s", stack.fqn, provider.get_delete_failed_status_reason(stack.fqn)
            )
            return FailedStatus(provider.get_stack_status_reason(stack_data))
        except CancelExecution:
            return SkippedStatus(reason="canceled execution")

    # TODO refactor long if, elif, else block
    # pylint: disable=too-many-return-statements,too-many-branches,too-many-statements
    def _launch_stack(self, stack: Stack, *, status: Status, **_: Any) -> Status:
        """Handle the creating or updating of a stack in CloudFormation.

        Also makes sure that we don't try to create or update a stack while
        it is already updating or creating.

        Args:
            stack: Stack to be launched.
            status: The Stack's status represented by a CFNgin status object.

        """
        wait_time = 0 if status is PENDING else STACK_POLL_TIME
        if self.cancel.wait(wait_time):
            return INTERRUPTED

        if not should_submit(stack):
            return NotSubmittedStatus()

        provider = self.build_provider()

        try:
            provider_stack = provider.get_stack(stack.fqn)
        except StackDoesNotExist:
            provider_stack = None

        if provider_stack and not should_update(stack):
            stack.set_outputs(self.provider.get_output_dict(provider_stack))
            return NotUpdatedStatus()

        recreate = False
        if provider_stack and status == SUBMITTED:
            LOGGER.debug(
                "%s:provider status: %s",
                stack.fqn,
                provider.get_stack_status(provider_stack),
            )

            if provider.is_stack_rolling_back(  # pylint: disable=no-else-return
                provider_stack
            ):
                if status.reason and "rolling back" in status.reason:
                    return status

                LOGGER.debug("%s:entered roll back", stack.fqn)
                if status.reason and "updating" in status.reason:
                    reason = "rolling back update"
                else:
                    reason = "rolling back new stack"

                return SubmittedStatus(reason)
            elif provider.is_stack_in_progress(provider_stack):
                LOGGER.debug("%s:in progress", stack.fqn)
                return status
            elif provider.is_stack_destroyed(provider_stack):
                LOGGER.debug("%s:finished deleting", stack.fqn)
                recreate = True
                # Continue with creation afterwards
            # Failure must be checked *before* completion, as both will be true
            # when completing a rollback, and we don't want to consider it as
            # a successful update.
            elif provider.is_stack_failed(provider_stack):
                reason = status.reason
                if reason and "rolling" in reason:
                    reason = reason.replace("rolling", "rolled")
                status_reason = provider.get_rollback_status_reason(stack.fqn)
                LOGGER.info("%s:roll back reason: %s", stack.fqn, status_reason)
                return FailedStatus(reason)

            elif provider.is_stack_completed(provider_stack):
                stack.set_outputs(provider.get_output_dict(provider_stack))
                return CompleteStatus(status.reason)
            else:
                return status

        LOGGER.debug("%s:resolving stack", stack.fqn)
        stack.resolve(self.context, self.provider)

        LOGGER.debug("%s:launching stack now", stack.fqn)
        template = self._template(stack.blueprint)
        stack_policy = self._stack_policy(stack)
        tags = build_stack_tags(stack)
        parameters = self.build_parameters(stack, provider_stack)
        force_change_set = stack.blueprint.requires_change_set

        if recreate:
            LOGGER.debug("%s:re-creating stack", stack.fqn)
            provider.create_stack(
                stack.fqn,
                template,
                parameters,
                tags,
                stack_policy=stack_policy,
                termination_protection=stack.termination_protection,
                timeout=stack.definition.timeout,
            )
            return SubmittedStatus("re-creating stack")
        if not provider_stack:
            LOGGER.debug("%s:creating new stack", stack.fqn)
            provider.create_stack(
                stack.fqn,
                template,
                parameters,
                tags,
                force_change_set=force_change_set,
                stack_policy=stack_policy,
                termination_protection=stack.termination_protection,
                timeout=stack.definition.timeout,
            )
            return SubmittedStatus("creating new stack")

        try:
            wait = stack.in_progress_behavior == "wait"
            if wait and provider.is_stack_in_progress(provider_stack):
                return WAITING
            if provider.prepare_stack_for_update(provider_stack, tags):
                existing_params = provider_stack.get("Parameters", [])
                provider.update_stack(
                    stack.fqn,
                    template,
                    existing_params,
                    parameters,
                    tags,
                    force_interactive=stack.protected,
                    force_change_set=force_change_set,
                    stack_policy=stack_policy,
                    termination_protection=stack.termination_protection,
                )

                LOGGER.debug("%s:updating existing stack", stack.fqn)
                return SubmittedStatus("updating existing stack")
            return SubmittedStatus("destroying stack for re-creation")
        except CancelExecution:
            stack.set_outputs(provider.get_output_dict(provider_stack))
            return SkippedStatus(reason="canceled execution")
        except StackDidNotChange:
            stack.set_outputs(provider.get_output_dict(provider_stack))
            return DidNotChangeStatus()

    @property
    def _stack_action(self) -> Callable[..., Status]:
        """Run against a step."""
        return self._launch_stack

    def _template(self, blueprint: Blueprint) -> Template:
        """Generate a template based on whether or not an S3 bucket is set.

        If an S3 bucket is set, then the template will be uploaded to S3 first,
        and CreateStack/UpdateStack operations will use the uploaded template.
        If not bucket is set, then the template will be inlined.

        """
        if self.upload_disabled:
            return Template(body=blueprint.rendered)
        return Template(url=self.s3_stack_push(blueprint))

    @staticmethod
    def _stack_policy(stack: Stack) -> Optional[Template]:
        """Return a Template object for the stacks stack policy."""
        return Template(body=stack.stack_policy) if stack.stack_policy else None

    def __generate_plan(self, tail: bool = False) -> Plan:
        """Plan creation that is specific to the build action.

        If a persistent graph is used, stacks that exist in the persistent
        graph but are no longer in the "local" graph will be deleted.
        If not using a persistent graph. the default method for creating
        a plan is used.

        Args:
            tail: Whether to tail the stack progress.

        """
        if not self.context.persistent_graph:
            return self._generate_plan(tail)

        graph = Graph()
        config_stack_names = [stack.name for stack in self.context.stacks]
        inverse_steps: List[Step] = []
        persist_graph = self.context.persistent_graph.transposed()

        for ind_node, dep_nodes in persist_graph.dag.graph.items():
            if ind_node not in config_stack_names:
                inverse_steps.append(
                    Step.from_stack_name(
                        ind_node,
                        self.context,
                        requires=list(dep_nodes),
                        fn=self._destroy_stack,
                        watch_func=(self._tail_stack if tail else None),
                    )
                )

        graph.add_steps(inverse_steps)

        # invert what is going to be destroyed to retain dependencies
        graph = graph.transposed()

        steps = [
            Step(
                stack,
                fn=self._launch_stack,
                watch_func=(self._tail_stack if tail else None),
            )
            for stack in self.context.stacks
        ]

        graph.add_steps(steps)

        return Plan(context=self.context, description=self.DESCRIPTION, graph=graph)

    def pre_run(
        self, *, dump: Union[bool, str] = False, outline: bool = False, **_: Any
    ) -> None:
        """Any steps that need to be taken prior to running the action."""
        if should_ensure_cfn_bucket(outline, bool(dump)):
            self.ensure_cfn_bucket()
        handle_hooks(
            "pre_deploy",
            self.context.config.pre_deploy,
            self.provider,
            self.context,
            dump=bool(dump),
            outline=outline,
        )

    def run(
        self,
        *,
        concurrency: int = 0,
        dump: Union[bool, str] = False,
        force: bool = False,  # pylint: disable=unused-argument
        outline: bool = False,
        tail: bool = False,
        upload_disabled: bool = False,
        **_kwargs: Any,
    ) -> None:
        """Kicks off the create/update of the stacks in the stack_definitions.

        This is the main entry point for the action.

        Args:
            concurrency: The maximum number of concurrent deployments.
            dump: Dump the plan rather than execute it.
            force: Not used by this action.
            outline: Outline the plan rather than execute it.
            tail: Tail the stack's events.
            upload_disabled: Whether to explicitly disable uploading the CloudFormation
                template to S3.

        """
        if upload_disabled:
            self.upload_disabled = upload_disabled
        plan = self.__generate_plan(tail=tail)
        if not plan.keys():
            LOGGER.warning("no stacks detected (error in config?)")
        if not outline and not dump:
            plan.outline(logging.DEBUG)
            self.context.lock_persistent_graph(plan.lock_code)
            LOGGER.debug("launching stacks: %s", ", ".join(plan.keys()))
            walker = build_walker(concurrency)
            try:
                plan.execute(walker)
            finally:
                # always unlock the graph at the end
                self.context.unlock_persistent_graph(plan.lock_code)
        if outline:
            plan.outline()
        if isinstance(dump, str):
            plan.dump(directory=dump, context=self.context, provider=self.provider)

    def post_run(
        self, *, dump: Union[bool, str] = False, outline: bool = False, **_: Any
    ) -> None:
        """Any steps that need to be taken after running the action."""
        handle_hooks(
            "post_deploy",
            self.context.config.post_deploy,
            self.provider,
            self.context,
            dump=bool(dump),
            outline=outline,
        )



"""Size only strategy.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/syncstrategy/sizeonly.py

"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar, Optional

from typing_extensions import Literal

from .base import BaseSync

if TYPE_CHECKING:
    from ..file_generator import FileStats


LOGGER = logging.getLogger(__name__.replace("._", "."))


class SizeOnlySync(BaseSync):
    """Only check size when determining if file should sync."""

    NAME: ClassVar[Literal["size_only"]] = "size_only"

    def determine_should_sync(
        self, src_file: Optional[FileStats], dest_file: Optional[FileStats]
    ) -> bool:
        """Determine if file should sync."""
        same_size = self.compare_size(src_file, dest_file)
        should_sync = not same_size
        if should_sync:
            LOGGER.debug(
                "syncing: %s -> %s, size_changed: %s",
                src_file.src if src_file else None,
                src_file.dest if src_file else None,
                not same_size,
            )
        return should_sync



"""S3 sync handler."""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Union

from .....compat import cached_property
from ._helpers.action_architecture import ActionArchitecture
from ._helpers.parameters import Parameters, ParametersDataModel
from ._helpers.sync_strategy.register import register_sync_strategies
from ._helpers.transfer_config import RuntimeConfig

if TYPE_CHECKING:
    import boto3
    from mypy_boto3_s3.client import S3Client

    from .....context import CfnginContext, RunwayContext
    from ._helpers.transfer_config import TransferConfigDict


class S3SyncHandler:
    """S3 sync handler."""

    def __init__(
        self,
        context: Union[CfnginContext, RunwayContext],
        *,
        delete: bool = False,
        dest: str,
        exclude: Optional[List[str]] = None,
        follow_symlinks: bool = False,
        include: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        session: Optional[boto3.Session] = None,
        src: str,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway or CFNgin context object.
            delete: If true, files that exist in the destination but not in the
                source are deleted.
            dest: Destination path.
            exclude: List of patterns for files/objects to exclude.
            follow_symlinks: If symlinks should be followed.
            include: List of patterns for files/objects to explicitly include.
            page_size: Number of items per page.
            session: boto3 Session.
            src: Source path.

        """
        self._session = session or context.get_session(region=context.env.aws_region)
        self._botocore_session = self._session._session
        self.ctx = context
        self.instructions = [
            "file_generator",
            "comparator",
            "file_info_builder",
            "s3_handler",
        ]
        self.parameters = Parameters(
            "sync",
            ParametersDataModel(
                delete=delete,
                dest=dest,
                exclude=exclude or [],
                follow_symlinks=follow_symlinks,
                include=include or [],
                page_size=page_size,
                src=src,
            ),
        )

    @cached_property
    def client(self) -> S3Client:
        """S3 client."""
        return self._session.client("s3")

    @cached_property
    def transfer_config(self) -> TransferConfigDict:
        """Get runtime transfer config."""
        return RuntimeConfig.build_config(
            **self._botocore_session.get_scoped_config().get("s3", {})
        )

    def run(self) -> None:
        """Run sync."""
        register_sync_strategies(self._botocore_session)  # type: ignore
        ActionArchitecture(
            session=self._session,
            botocore_session=self._botocore_session,  # type: ignore
            action="sync",
            parameters=self.parameters.data,
            runtime_config=self.transfer_config,
        ).run()



"""CFNgin prehook responsible for creation of Lambda@Edge functions."""
from __future__ import annotations

import logging
import os
import re
import secrets
import shutil
import tempfile
from tempfile import mkstemp
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from ... import aws_lambda
from ...base import HookArgsBaseModel

if TYPE_CHECKING:
    from .....context import CfnginContext
    from ....providers.aws.default import Provider

# The functions associated with Auth@Edge
FUNCTIONS = ["check_auth", "refresh_auth", "parse_auth", "sign_out", "http_headers"]


LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    bucket: str
    """S3 bucket name."""

    client_id: str
    """The ID of the Cognito User Pool Client."""

    cookie_settings: Dict[str, Any]
    """The settings for our customized cookies."""

    http_headers: Dict[str, Any]
    """The additional headers added to our requests."""

    nonce_signing_secret_param_name: str
    """SSM param name to store nonce signing secret."""

    oauth_scopes: List[str]
    """The validation scopes for our OAuth requests."""

    redirect_path_refresh: str
    """The URL path for authorization refresh redirect (Correlates to the refresh auth lambda)."""

    redirect_path_sign_in: str
    """The URL path to be redirected to after sign in (Correlates to the parse auth lambda)."""

    redirect_path_sign_out: str
    """The URL path to be redirected to after sign out (Correlates to the root to
    be asked to resigning).

    """

    required_group: Optional[str] = None
    """Optional User Pool group to which access should be restricted."""


# pylint: disable=too-many-locals
def write(
    context: CfnginContext, provider: Provider, *__args: Any, **kwargs: Any
) -> Dict[str, Any]:
    """Writes/Uploads the configured lambdas for Auth@Edge.

    Lambda@Edge does not have the ability to allow Environment variables
    at the time of this writing. In order to configure our lambdas with
    dynamic variables we first will go through and update a "shared" template
    with all of the configuration elements and add that to a temporary
    folder along with each of the individual Lambda@Edge functions. This
    temporary folder is then used with the CFNgin awsLambda hook to build
    the functions.

    """
    cognito_domain = context.hook_data["aae_domain_updater"].get("domain")
    args = HookArgs.parse_obj(kwargs)
    config = {
        "client_id": args.client_id,
        "cognito_auth_domain": cognito_domain,
        "cookie_settings": args.cookie_settings,
        "http_headers": args.http_headers,
        "oauth_scopes": args.oauth_scopes,
        "redirect_path_auth_refresh": args.redirect_path_refresh,
        "redirect_path_sign_in": args.redirect_path_sign_in,
        "redirect_path_sign_out": args.redirect_path_sign_out,
        "required_group": args.required_group,
        "user_pool_id": context.hook_data["aae_user_pool_id_retriever"]["id"],
        "nonce_signing_secret": get_nonce_signing_secret(
            args.nonce_signing_secret_param_name, context
        ),
    }

    # Shared file that contains the method called for configuration data
    path = os.path.join(os.path.dirname(__file__), "templates", "shared.py")
    context_dict: Dict[str, Any] = {}

    with open(path, encoding="utf-8") as file_:
        # Dynamically replace our configuration values
        # in the shared.py template file with actual
        # calculated values
        shared = re.sub(
            r"{.+?(})$", str(config), file_.read(), 1, flags=re.DOTALL | re.MULTILINE
        )

        filedir, temppath = mkstemp()

        # Save the file to a temp path
        with open(temppath, "w", encoding="utf-8") as tmp:
            tmp.write(shared)
            config = temppath
        os.close(filedir)

        # Get all of the different Auth@Edge functions
        for handler in FUNCTIONS:
            # Create a temporary folder
            dirpath = tempfile.mkdtemp()

            # Copy the template code for the specific Lambda function
            # to the temporary folder
            shutil.copytree(
                os.path.join(os.path.dirname(__file__), "templates", handler),
                dirpath,
                dirs_exist_ok=True,
            )

            # Save our dynamic configuration shared file to the
            # temporary folder
            with open(config, encoding="utf-8") as shared:
                raw = shared.read()
                filename = "shared.py"
                with open(os.path.join(dirpath, filename), "wb") as newfile:
                    newfile.write(raw.encode())

            # Copy the shared jose-dependent util module to the temporary folder
            shutil.copyfile(
                os.path.join(os.path.dirname(__file__), "templates", "shared_jose.py"),
                os.path.join(dirpath, "shared_jose.py"),
            )

            # Upload our temporary folder to our S3 bucket for
            # Lambda use
            lamb = aws_lambda.upload_lambda_functions(
                context,
                provider,
                bucket=args.bucket,
                functions={
                    handler: {
                        "path": dirpath,
                        "python_dontwritebytecode": True,
                        "python_exclude_bin_dir": True,
                        "python_exclude_setuptools_dirs": True,
                    }
                },
            )

            # Add the lambda code reference to our context_dict
            context_dict.update(lamb)

    return context_dict


def get_nonce_signing_secret(param_name: str, context: CfnginContext) -> str:
    """Retrieve signing secret, generating & storing it first if not present."""
    session = context.get_session()
    ssm_client = session.client("ssm")
    try:
        response = ssm_client.get_parameter(Name=param_name, WithDecryption=True)
        return response["Parameter"]["Value"]
    except ssm_client.exceptions.ParameterNotFound:
        secret = random_key(16)
        ssm_client.put_parameter(
            Description="Auth@Edge nonce signing secret",
            Name=param_name,
            Type="String",
            Value=secret,
        )
        return secret


def random_key(length: int = 16) -> str:
    """Generate a random key of specified length from the allowed secret characters.

    Args:
        length: The length of the random key.

    """
    secret_allowed_chars = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~"
    )
    return "".join(secrets.choice(secret_allowed_chars) for _ in range(length))



"""Assume an AWS IAM role."""
from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, ContextManager, Optional, Type, cast

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from types import TracebackType

    from mypy_boto3_sts.type_defs import AssumedRoleUserTypeDef, CredentialsTypeDef

    from ...._logging import RunwayLogger
    from ....context import RunwayContext


LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))

_KwargsTypeDef = TypedDict(
    "_KwargsTypeDef", DurationSeconds=int, RoleArn=str, RoleSessionName=str
)


class AssumeRole(ContextManager["AssumeRole"]):
    """Context manager for assuming an AWS role."""

    assumed_role_user: AssumedRoleUserTypeDef
    credentials: CredentialsTypeDef
    ctx: RunwayContext
    duration_seconds: int
    revert_on_exit: bool
    session_name: str = "runway"

    def __init__(
        self,
        context: RunwayContext,
        role_arn: Optional[str] = None,
        duration_seconds: Optional[int] = None,
        revert_on_exit: bool = True,
        session_name: Optional[str] = None,
    ):
        """Instantiate class.

        Args:
            context: Runway context object.
            role_arn: ARN of role to be assumed.
            duration_seconds: Seconds that the assumed role's credentials will be
                valid for. (default: 3600)
            revert_on_exit: Whether credentials in the environment will be
                reverted upon exiting the context manager.
            session_name: Name to use for the assumed role session. (default: runway)

        """
        self.assumed_role_user = {"AssumedRoleId": "", "Arn": ""}
        self.credentials = {
            "AccessKeyId": "",
            "Expiration": datetime.now(),
            "SecretAccessKey": "",
            "SessionToken": "",
        }
        self.role_arn = role_arn
        self.ctx = context
        self.duration_seconds = duration_seconds or 3600
        self.revert_on_exit = revert_on_exit
        self.session_name = session_name or "runway"

    @property
    def _kwargs(self) -> _KwargsTypeDef:
        """Construct keyword arguments to pass to boto3 call."""
        return {
            "DurationSeconds": self.duration_seconds,
            "RoleArn": self.role_arn or "",
            "RoleSessionName": self.session_name,
        }

    def assume(self) -> None:
        """Perform role assumption."""
        if not self.role_arn:
            LOGGER.debug("no role to assume")
            return
        if self.revert_on_exit:
            self.save_existing_iam_env_vars()
        sts_client = self.ctx.get_session().client("sts")
        LOGGER.info("assuming role %s...", self.role_arn)
        response = sts_client.assume_role(**self._kwargs)
        LOGGER.debug("sts.assume_role respsone: %s", response)
        if "Credentials" in response:
            self.assumed_role_user.update(
                response.get("AssumedRoleUser", cast("AssumedRoleUserTypeDef", {}))
            )
            self.credentials.update(response["Credentials"])
            self.ctx.env.vars.update(
                {
                    "AWS_ACCESS_KEY_ID": response["Credentials"]["AccessKeyId"],
                    "AWS_SECRET_ACCESS_KEY": response["Credentials"]["SecretAccessKey"],
                    "AWS_SESSION_TOKEN": response["Credentials"]["SessionToken"],
                }
            )
            LOGGER.verbose("updated environment with assumed credentials")
        else:
            raise ValueError("assume_role did not return Credentials")

    def restore_existing_iam_env_vars(self) -> None:
        """Restore backed up IAM environment variables."""
        if not self.role_arn:
            LOGGER.debug("no role was assumed; not reverting credentials")
            return
        for k in self.ctx.current_aws_creds.keys():
            old = "OLD_" + k
            if self.ctx.env.vars.get(old):
                self.ctx.env.vars[k] = self.ctx.env.vars.pop(old)
                LOGGER.debug("reverted environment variables: %s", k)
            else:
                self.ctx.env.vars.pop(k, None)
                LOGGER.debug("removed environment variables: %s ", k)

    def save_existing_iam_env_vars(self) -> None:
        """Backup IAM environment variables for later restoration."""
        for k, v in self.ctx.current_aws_creds.items():
            new = "OLD_" + k
            LOGGER.debug('saving environment variable "%s" as "%s"', k, new)
            self.ctx.env.vars[new] = cast(str, v)

    def __enter__(self) -> AssumeRole:
        """Enter the context manager."""
        LOGGER.debug("entering aws.AssumeRole context manager...")
        self.assume()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the context manager."""
        if self.revert_on_exit:
            self.restore_existing_iam_env_vars()
        LOGGER.debug("aws.AssumeRole context manager exited")



"""Environment variable lookup."""
# pyright: reportIncompatibleMethodOverride=none
import logging
import os
from typing import Any

from typing_extensions import Final, Literal

from ....lookups.handlers.base import LookupHandler
from ...utils import read_value_from_path

LOGGER = logging.getLogger(__name__)


class EnvvarLookup(LookupHandler):
    """Environment variable lookup."""

    DEPRECATION_MSG = "envvar Lookup has been deprecated; use the env lookup instead"
    TYPE_NAME: Final[Literal["envvar"]] = "envvar"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, **_: Any) -> str:  # pylint: disable=arguments-differ
        """Retrieve an environment variable.

        Args:
            value: Parameter(s) given to this lookup.

        Example:
            ::

                # With CFNgin we would reference the environment variable like this:
                conf_key: ${envvar ENV_VAR_NAME}

            You can optionally store the value in a file, ie::

                $ cat envvar_value.txt
                ENV_VAR_NAME

            and reference it within CFNgin (NOTE: the path should be relative
            to the CFNgin config file)::

                conf_key: ${envvar file://envvar_value.txt}

                # Both of the above would resolve to
                conf_key: ENV_VALUE

        """
        LOGGER.warning("${envvar %s}: %s: ${env %s}", value, cls.DEPRECATION_MSG, value)
        value = read_value_from_path(value)

        try:
            return os.environ[value]
        except KeyError as exc:
            raise ValueError(f'EnvVar "{value}" does not exist') from exc



"""``runway plan`` command."""
# docs: file://./../../../docs/source/commands.rst
import logging
import os
import platform
from typing import TYPE_CHECKING, Any, Dict, cast

import click
from pydantic import ValidationError

from ...core import Runway
from ...exceptions import ConfigNotFound, VariablesFileNotFound
from .. import options

if TYPE_CHECKING:
    from ..._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("envvars", short_help="exportable env_vars")
@options.debug
@options.deploy_environment
@options.no_color
@options.verbose
@click.pass_context
def envvars(ctx: click.Context, debug: bool, **_: Any) -> None:
    """Output "env_vars" defined in the Runway config file.

    OS environment variables can be set in the Runway config file for different
    Runway environments (e.g. dev & prod KUBECONFIG values).
    This command allows access to these values for use outside of Runway.

    NOTE: Only outputs "env_vars" defined in deployments, not modules.

    """
    if not (ctx.obj.debug or ctx.obj.verbose):
        logging.getLogger("runway").setLevel(logging.ERROR)  # suppress warnings

    ctx.obj.env.ci = True
    LOGGER.verbose("forced Runway to non-interactive mode to suppress prompts")
    try:
        env_vars = Runway(
            ctx.obj.runway_config, ctx.obj.get_runway_context()
        ).get_env_vars()
    except ValidationError as err:
        LOGGER.error(err, exc_info=debug)
        ctx.exit(1)
    except (ConfigNotFound, VariablesFileNotFound) as err:
        LOGGER.error(err.message, exc_info=debug)
        ctx.exit(1)

    if not env_vars:
        LOGGER.error("No env_vars defined in %s", ctx.obj.runway_config_path)
        ctx.exit(1)
    LOGGER.debug("env_vars: %s", env_vars)
    print_env_vars(env_vars)


def print_env_vars(env_vars: Dict[str, Any]) -> None:
    """Print environment variables."""
    if platform.system() == "Windows":
        if os.getenv("MSYSTEM", "").startswith("MINGW"):
            return __print_env_vars_posix(env_vars)  # git bash
        return __print_env_vars_psh(env_vars)
    return __print_env_vars_posix(env_vars)


def __print_env_vars_posix(env_vars: Dict[str, Any]) -> None:
    """Print environment variables for bash."""
    LOGGER.debug("using posix formating for environment variable export")
    for key, val in env_vars.items():
        click.echo(f'export {key}="{val}"')


def __print_env_vars_psh(env_vars: Dict[str, Any]) -> None:
    """Print environment variables for Powershell."""
    LOGGER.debug("using powershell formating for environment variable export")
    for key, val in env_vars.items():
        click.echo(f'$env:{key} = "{val}"')



"""``runway gen-sample k8s-cfn`` command."""
import locale
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click
from cfn_flip import to_yaml

from ....blueprints.k8s.k8s_iam import Iam
from ....blueprints.k8s.k8s_master import Cluster
from ....blueprints.k8s.k8s_workers import NodeGroup
from ....context import CfnginContext
from ... import options
from .utils import TEMPLATES, convert_gitignore, copy_sample

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("k8s-cfn-repo", short_help="k8s + cfn (k8s-cfn-infrastructure)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def k8s_cfn_repo(ctx: click.Context, **_: Any) -> None:
    """Generate a sample CloudFormation project using Kubernetes."""
    src = TEMPLATES / "k8s-cfn-repo"
    dest = Path.cwd() / "k8s-cfn-infrastructure"

    copy_sample(ctx, src, dest)
    convert_gitignore(dest / "_gitignore")

    master_templates = dest / "k8s-master.cfn/templates"
    worker_templates = dest / "k8s-workers.cfn/templates"
    env = {"namespace": "test"}

    LOGGER.verbose("rendering master templates...")
    encoding = locale.getpreferredencoding(do_setlocale=False)
    master_templates.mkdir()
    (master_templates / "k8s_iam.yaml").write_text(
        to_yaml(Iam("test", CfnginContext(environment=env.copy())).to_json()),
        encoding=encoding,
    )
    (master_templates / "k8s_master.yaml").write_text(
        to_yaml(Cluster("test", CfnginContext(environment=env.copy())).to_json()),
        encoding=encoding,
    )

    LOGGER.verbose("rendering worker templates...")
    worker_templates.mkdir()
    (worker_templates / "k8s_workers.yaml").write_text(
        to_yaml(NodeGroup("test", CfnginContext(environment=env.copy())).to_json()),
        encoding=encoding,
    )

    LOGGER.success("Sample k8s infrastructure repo created at %s", dest)
    LOGGER.notice("See the README for setup and deployment instructions.")



"""AWS S3 bucket."""
from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Any, List, Optional, Union

from botocore.exceptions import ClientError

from .....compat import cached_property
from .....mixins import DelCachedPropMixin
from .._response import BaseResponse
from ._sync_handler import S3SyncHandler

if TYPE_CHECKING:
    import boto3
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import (
        CreateBucketOutputTypeDef,
        GetBucketVersioningOutputTypeDef,
    )

    from .....context import CfnginContext, RunwayContext

LOGGER = logging.getLogger(__name__.replace("._", "."))


class Bucket(DelCachedPropMixin):
    """AWS S3 bucket."""

    def __init__(
        self,
        context: Union[CfnginContext, RunwayContext],
        name: str,
        region: Optional[str] = None,
    ) -> None:
        """Instantiate class.

        Args:
            context: Current context object.
            name: The name of the bucket.
            region: The bucket's region.

        """
        self.__ctx = context
        self._region = region
        self.name = name

    @cached_property
    def client(self) -> S3Client:
        """Create or reuse a boto3 client."""
        return self.session.client("s3")

    @property
    def exists(self) -> bool:
        """Check whether the bucket exists and is accessible."""
        return not self.forbidden and not self.not_found

    @property
    def forbidden(self) -> bool:
        """Check whether access to the bucket is forbidden."""
        return self.head.metadata.forbidden

    @cached_property
    def head(self):
        """Check if a bucket exists and you have permission to access it.

        To use this operation, the user must have permissions to perform the
        ``s3:ListBucket`` action.

        This is a low level action that returns the raw result of the request.

        """
        try:
            return BaseResponse(**self.client.head_bucket(Bucket=self.name) or {})
        except ClientError as err:
            LOGGER.debug(
                'received an error from AWS S3 when trying to head bucket "%s"',
                self.name,
                exc_info=True,
            )
            return BaseResponse.parse_obj(err.response)

    @property
    def not_found(self) -> bool:
        """Check whether the bucket exists."""
        return self.head.metadata.not_found

    @cached_property
    def session(self) -> boto3.Session:
        """Create cached boto3 session."""
        return self.__ctx.get_session(region=self._region)

    def create(self, **kwargs: Any) -> Optional[CreateBucketOutputTypeDef]:
        """Create an S3 Bucket if it does not already exist.

        Bucket creation will be skipped if it already exists or access is forbidden.

        Keyword arguments are passed directly to the boto3 method.

        Returns:
            boto3 response.

        """
        if self.forbidden or self.exists:
            LOGGER.debug(
                'skipped creating bucket "%s": %s',
                self.name,
                "access denied" if self.forbidden else "bucket already exists",
            )
            return None
        kwargs["Bucket"] = self.name
        if self.client.meta.region_name != "us-east-1":
            kwargs.setdefault("CreateBucketConfiguration", {})
            kwargs["CreateBucketConfiguration"].update(
                {"LocationConstraint": self.client.meta.region_name}
            )
        LOGGER.debug("creating bucket: %s", json.dumps(kwargs))
        self._del_cached_property("head")
        return self.client.create_bucket(**kwargs)

    def enable_versioning(self) -> None:
        """Enable versioning on the bucket if not already enabled."""
        config = self.get_versioning()
        if config.get("Status") == "Enabled":
            LOGGER.debug(
                'did not modify versioning policy for bucket "%s"; already enabled',
                self.name,
            )
            return
        config["Status"] = "Enabled"
        self.client.put_bucket_versioning(
            Bucket=self.name,
            VersioningConfiguration={
                "Status": "Enabled",
                "MFADelete": config.get("MFADelete", "Disabled"),
            },
        )
        LOGGER.debug('enabled versioning for bucket "%s"', self.name)

    def format_bucket_path_uri(
        self, *, key: Optional[str] = None, prefix: Optional[str] = None
    ) -> str:
        """Format bucket path URI.

        Args:
            key: S3 object key.
            prefix: Directory tree to append to key.

        Returns:
            S3 bucket URI in ``s3://{bucket-name}/{prefix}/{key}`` format

        """
        uri = f"s3://{self.name}"
        if prefix:
            uri += f"/{prefix}"
        if key:
            uri += f"/{key}"
        return uri

    def get_versioning(self) -> GetBucketVersioningOutputTypeDef:
        """Get the versioning state of a bucket.

        To retrieve the versioning state of a bucket, you must be the bucket owner.

        Returns:
            The current versioning state of the bucket containing ``Status``
            and ``MFADelete`` (only if this has ever been configured).

        """
        return self.client.get_bucket_versioning(Bucket=self.name)

    def sync_from_local(
        self,
        src_directory: str,
        *,
        delete: bool = False,
        exclude: Optional[List[str]] = None,
        follow_symlinks: bool = False,
        include: Optional[List[str]] = None,
        prefix: Optional[str] = None,
    ) -> None:
        """Sync local directory to the S3 Bucket.

        Args:
            src_directory: Local directory to sync to S3.
            delete: If true, files that exist in the destination but not in the
                source are deleted.
            exclude: List of patterns for files/objects to exclude.
            follow_symlinks: If symlinks should be followed.
            include: List of patterns for files/objects to explicitly include.
            prefix: Optional prefix to append to synced objects.

        """
        S3SyncHandler(
            context=self.__ctx,
            delete=delete,
            dest=self.format_bucket_path_uri(prefix=prefix),
            exclude=exclude,
            follow_symlinks=follow_symlinks,
            include=include,
            session=self.session,
            src=src_directory,
        ).run()

    def sync_to_local(
        self,
        dest_directory: str,
        *,
        delete: bool = False,
        exclude: Optional[List[str]] = None,
        follow_symlinks: bool = False,
        include: Optional[List[str]] = None,
        prefix: Optional[str] = None,
    ) -> None:
        """Sync S3 bucket to local directory.

        Args:
            dest_directory: Local directory to sync S3 objects to.
            delete: If true, files that exist in the destination but not in the
                source are deleted.
            exclude: List of patterns for files/objects to exclude.
            follow_symlinks: If symlinks should be followed.
            include: List of patterns for files/objects to explicitly include.
            prefix: Optional prefix to append to synced objects.

        """
        S3SyncHandler(
            context=self.__ctx,
            delete=delete,
            dest=dest_directory,
            exclude=exclude,
            follow_symlinks=follow_symlinks,
            include=include,
            session=self.session,
            src=self.format_bucket_path_uri(prefix=prefix),
        ).run()

    def __bool__(self) -> bool:
        """Implement evaluation of instances as a bool."""
        return self.exists



"""Runway module utilities."""
from __future__ import annotations

import logging
import os
import platform
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union, cast

from ..utils import which

if TYPE_CHECKING:
    from typing import Any

    from .._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))
NPM_BIN = "npm.cmd" if platform.system().lower() == "windows" else "npm"
NPX_BIN = "npx.cmd" if platform.system().lower() == "windows" else "npx"


def format_npm_command_for_logging(command: List[str]) -> str:
    """Convert npm command list to string for display to user."""
    if platform.system().lower() == "windows" and (
        command[0] == "npx.cmd" and command[1] == "-c"
    ):
        return f'npx.cmd -c "{" ".join(command[2:])}"'
    return " ".join(command)


# type hint quoted b/c pylint 2.11.1 raises unsubscriptable-object
def generate_node_command(
    command: str,
    command_opts: List[str],
    path: Path,
    *,
    logger: Union[logging.Logger, "logging.LoggerAdapter[Any]"] = LOGGER,
    package: Optional[str] = None,
) -> List[str]:
    """Return node bin command list for subprocess execution.

    Args:
        command: Command to execute from a local ``node_modules/.bin``.
        command_opts: Options to include with the command.
        path: Current working directory. Used to construct a "fall-back" command
            when ``npx`` is not available/included as part of npm.
        logger: A specific logger to use when logging the constructed command.
        package: Name of the npm package containing the binary to execute.
            This is recommended when the name of the binary does not match the
            name of the npm package.

    """
    if which(NPX_BIN):
        # Use npx if available (npm v5.2+)
        cmd_list = [NPX_BIN]
        if package:
            cmd_list.extend(
                [
                    "--package",
                    package,
                    command,
                    *command_opts,
                ]
            )
        else:
            cmd_list.append("-c")
            cmd_list.append(f"{command} {' '.join(command_opts)}".strip())
    else:
        logger.debug("npx not found; falling back to invoking shell script directly")
        cmd_list = [str(path / "node_modules" / ".bin" / command), *command_opts]
    logger.debug("node command: %s", format_npm_command_for_logging(cmd_list))
    return cmd_list


# type hint quated b/c pylint 2.11.1 raises unsubscriptable-object
def run_module_command(
    cmd_list: List[str],
    env_vars: Dict[str, str],
    exit_on_error: bool = True,
    logger: Union[logging.Logger, "logging.LoggerAdapter[Any]"] = LOGGER,
) -> None:
    """Shell out to provisioner command.

    Args:
        cmd_list: Command to run.
        env_vars: Environment variables.
        exit_on_error: If true, ``subprocess.CalledProcessError`` will be caught
            and the resulting exit code will be passed to ``sys.exit()``.
            If false, the error will not be caught within this function.
            logger: Optionally, supply a logger to use.
        logger: A specific logger to use when logging the constructed command.

    """
    logger.debug("running command: %s", " ".join(cmd_list))
    if exit_on_error:
        try:
            subprocess.check_call(cmd_list, env=env_vars)
        except subprocess.CalledProcessError as shelloutexc:
            sys.exit(shelloutexc.returncode)
    else:
        subprocess.check_call(cmd_list, env=env_vars)


def use_npm_ci(path: Path) -> bool:
    """Return true if npm ci should be used in lieu of npm install."""
    # https://docs.npmjs.com/cli/ci#description
    with open(os.devnull, "w", encoding="utf-8") as fnull:
        if (
            (path / "package-lock.json").is_file()
            or (path / "npm-shrinkwrap.json").is_file()
        ) and subprocess.call(
            [NPM_BIN, "ci", "-h"], stdout=fnull, stderr=subprocess.STDOUT
        ) == 0:
            return True
    return False



"""AWS CloudFormation Output lookup."""
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any, Dict, NamedTuple, Set, Tuple

from typing_extensions import Final, Literal

from ....exceptions import OutputDoesNotExist
from ....lookups.handlers.base import LookupHandler
from ....utils import DOC_SITE
from ...exceptions import StackDoesNotExist

if TYPE_CHECKING:
    from ....context import CfnginContext
    from ....variables import VariableValue

LOGGER = logging.getLogger(__name__)


class OutputQuery(NamedTuple):
    """Output query NamedTuple."""

    stack_name: str
    output_name: str


class OutputLookup(LookupHandler):
    """AWS CloudFormation Output lookup."""

    DEPRECATION_MSG = (
        'lookup query syntax "<relative-stack-name>::<OutputName>" has been deprecated; '
        "to learn how to use the new lookup query syntax visit "
        f"{DOC_SITE}/page/cfngin/lookups/output.html"
    )
    TYPE_NAME: Final[Literal["output"]] = "output"
    """Name that the Lookup is registered as."""

    @classmethod
    def legacy_parse(cls, value: str) -> Tuple[OutputQuery, Dict[str, str]]:
        """Retain support for legacy lookup syntax.

        Format of value:
            <relative-stack-name>::<OutputName>

        """
        LOGGER.warning("${%s %s}: %s", cls.TYPE_NAME, value, cls.DEPRECATION_MSG)
        return deconstruct(value), {}

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls, value: str, context: CfnginContext, **_: Any
    ) -> str:
        """Fetch an output from the designated stack.

        Args:
            value: Parameter(s) given to this lookup.
                ``<relative-stack-name>.<OutputName>``
            context: Context instance.

        Returns:
            Output from the specified stack.

        Raises:
            OutputDoesNotExist: Output not found for Stack.
            StackDoesNotExist: Stack not found for the name provided.

        """
        try:
            raw_query, args = cls.parse(value)
            query = OutputQuery(*raw_query.split("."))
        except ValueError:
            query, args = cls.legacy_parse(value)

        stack = context.get_stack(query.stack_name)
        if not stack:
            if "default" in args:
                return cls.format_results(args["default"], **args)
            raise StackDoesNotExist(context.get_fqn(query.stack_name))

        if "default" in args:  # handle falsy default
            return cls.format_results(
                stack.outputs.get(query.output_name, args["default"]), **args
            )

        try:
            return cls.format_results(stack.outputs[query.output_name], **args)
        except KeyError:
            raise OutputDoesNotExist(
                stack_name=context.get_fqn(query.stack_name), output=query.output_name
            ) from None

    @classmethod
    def dependencies(cls, lookup_query: VariableValue) -> Set[str]:
        """Calculate any dependencies required to perform this lookup.

        Note that lookup_query may not be (completely) resolved at this time.

        Args:
            lookup_query: Parameter(s) given to this lookup.

        Returns:
            Stack names this lookup depends on.

        """
        # try to get the stack name
        stack_name = ""
        for data_item in lookup_query:
            if not data_item.resolved:
                # We encountered an unresolved substitution.
                # StackName is calculated dynamically based on context:
                #  e.g. ${output ${default var::source}.name}
                # Stop here
                return set()
            stack_name += data_item.value
            match = re.search(r"(::|\.)", stack_name)
            if match:
                stack_name = stack_name[0 : match.start()]
                return {stack_name}
                # else: try to append the next item

        # We added all lookup_query, and still couldn't find a `::`...
        # Probably an error...
        return set()


def deconstruct(value: str) -> OutputQuery:  # TODO remove in next major release
    """Deconstruct the value."""
    try:
        stack_name, output_name = value.split("::")
    except ValueError:
        raise ValueError(
            f"output handler requires syntax of <stack>::<output>. Got: {value}"
        ) from None

    return OutputQuery(stack_name, output_name)



"""File filter.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/filters.py

"""
from __future__ import annotations

import fnmatch
import logging
import os
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generator,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    cast,
)

from typing_extensions import Literal

from .utils import split_s3_bucket_key

if TYPE_CHECKING:
    from .file_generator import FileStats
    from .parameters import ParametersDataModel

LOGGER = logging.getLogger(__name__.replace("._", "."))


_FilterType = Literal["exclude", "include"]
FileStatus = NamedTuple("FileStatus", [("file_stats", "FileStats"), ("include", bool)])
FilterPattern = NamedTuple("FilterPattern", [("type", _FilterType), ("pattern", str)])


class Filter:
    """Universal exclude/include filter."""

    FILTER_TYPES: ClassVar[Tuple[_FilterType, ...]] = (
        "exclude",
        "include",
    )

    def __init__(
        self,
        patterns: Iterable[FilterPattern],
        src_rootdir: Optional[str],
        dest_rootdir: Optional[str],
    ) -> None:
        """Instantiate class.

        Args:
            patterns: List of filter patterns.
            src_rootdir: The source root directory where the patterns are
                evaluated.
            dest_rootdir: The destination root directory where the patterns are
                evaluated.

        """
        self._original_patterns = patterns
        self.patterns = self._full_path_patterns(patterns, src_rootdir)
        self.dest_patterns = self._full_path_patterns(patterns, dest_rootdir)

    @staticmethod
    def _full_path_patterns(
        patterns: Iterable[FilterPattern], rootdir: Optional[str]
    ) -> List[FilterPattern]:
        """Prefix each pattern with the root directory.

        Args:
            patterns: List of filter patterns.
            rootdir: The root directory to add to each pattern.

        """
        if rootdir:
            return sorted(  # sort for consistency
                [
                    FilterPattern(type=f.type, pattern=os.path.join(rootdir, f.pattern))
                    for f in patterns
                ]
            )
        return list(patterns)

    def call(self, files: Iterator[FileStats]) -> Generator[FileStats, None, None]:
        """Iterate over the yielded FileStats objects.

        Determines the type of the file and applies pattern matching to
        determine if the rule applies. While iterating though the patterns the
        file is assigned a boolean flag to determine if a file should be
        yielded on past the filer. Anything identified by the exclude filter
        has its flag set to false. Anything identified by the include filter
        has its flag set to True. All files begin with the flag set to true.
        Rules listed at the end will overwrite flags thrown by rules listed
        before it.

        """
        for file_stats in files:
            file_status = FileStatus(file_stats, True)
            for src_fp, dest_fp in zip(self.patterns, self.dest_patterns):
                current_file_status = self._match_pattern(src_fp, file_stats)
                if current_file_status is not None:
                    file_status = current_file_status
                dst_current_file_status = self._match_pattern(dest_fp, file_stats)
                if dst_current_file_status is not None:
                    file_status = dst_current_file_status
            LOGGER.debug(
                "=%s final filtered status, should_include: %s",
                file_stats.src,
                file_status.include,
            )
            if file_status.include:
                yield file_stats

    @staticmethod
    def _match_pattern(
        filter_pattern: FilterPattern, file_stats: FileStats
    ) -> Optional[FileStatus]:
        """Match file to pattern.

        Args:
            filter_pattern: Filter pattern to run against file.
            file_stats: Information about a file.

        """
        file_status = None
        if file_stats.src_type == "local":
            path_pattern = filter_pattern.pattern.replace("/", os.sep)
        else:
            path_pattern = filter_pattern.pattern.replace(os.sep, "/")
        is_match = fnmatch.fnmatch(str(file_stats.src), path_pattern)
        if is_match and filter_pattern.type == "include":
            file_status = FileStatus(file_stats, True)
            LOGGER.debug("%s matched include filter: %s", file_stats.src, path_pattern)
        elif is_match and filter_pattern.type == "exclude":
            file_status = FileStatus(file_stats, False)
            LOGGER.debug("%s matched exclude filter: %s", file_stats.src, path_pattern)
        else:
            LOGGER.debug(
                "%s did not match %s filter: %s",
                file_stats.src,
                filter_pattern.type,
                path_pattern,
            )
        return file_status

    @classmethod
    def parse_params(cls, parameters: ParametersDataModel) -> Filter:
        """Parse parameters to create a Filter instance."""
        if not (parameters.exclude or parameters.include):
            return Filter([], None, None)
        filter_patterns: Set[FilterPattern] = set()
        for filter_type in cls.FILTER_TYPES:
            for pat in parameters[filter_type]:
                filter_patterns.add(
                    FilterPattern(type=cast(_FilterType, filter_type), pattern=pat)
                )
        return Filter(
            filter_patterns,
            cls.parse_rootdir(parameters.src),
            cls.parse_rootdir(parameters.dest),
        )

    @classmethod
    def parse_rootdir(cls, path: str, dir_op: bool = True) -> str:
        """Parse path to extract the root directory.

        Args:
            path: Path to parse.
            dir_op: If the path is a directory.

        """
        if path.startswith("s3://"):
            return cls._parse_rootdir_s3(path, dir_op)
        return cls._parse_rootdir_local(path, dir_op)

    @staticmethod
    def _parse_rootdir_local(path: str, dir_op: bool = True) -> str:
        """Get root directory from local path.

        Args:
            path: Path to parse.
            dir_op: If the path is a directory.

        """
        if dir_op:
            return os.path.abspath(path)
        return os.path.abspath(os.path.dirname(path))

    @staticmethod
    def _parse_rootdir_s3(path: str, dir_op: bool = True) -> str:
        """Get root directory from S3 path.

        Args:
            path: Path to parse.
            dir_op: If the path is a directory.

        """
        bucket, key = split_s3_bucket_key(path)
        if not (dir_op or key.endswith("/")):
            key = "/".join(key.split("/")[:-1])
        return "/".join([bucket, key])



"""Base class for lookup handlers."""
from __future__ import annotations

import json
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

import yaml
from troposphere import BaseAWSObject
from typing_extensions import Literal

from ...cfngin.utils import read_value_from_path
from ...utils import MutableMap

if TYPE_CHECKING:
    from ...cfngin.providers.aws.default import Provider
    from ...context import CfnginContext, RunwayContext
    from ...variables import VariableValue

LOGGER = logging.getLogger(__name__)

TransformToTypeLiteral = Literal["bool", "str"]


def str2bool(v):  # type: ignore
    """Return boolean value of string."""
    return v.lower() in ("yes", "true", "t", "1", "on")


class LookupHandler:
    """Base class for lookup handlers."""

    TYPE_NAME: ClassVar[str]
    """Name that the Lookup is registered as."""

    @classmethod
    def dependencies(cls, __lookup_query: VariableValue) -> Set[str]:
        """Calculate any dependencies required to perform this lookup.

        Note that lookup_query may not be (completely) resolved at this time.

        """
        return set()

    @classmethod
    def format_results(
        cls,
        value: Any,
        get: Optional[str] = None,
        load: Optional[str] = None,
        transform: Optional[TransformToTypeLiteral] = None,
        **kwargs: Any,
    ) -> Any:
        """Format results to be returned by a lookup.

        Args:
            value: Data collected by the Lookup.
            get: Nested value to get from a dictionary like object.
            load: Parser to use to parse a formatted string before the ``get``
                and ``transform`` method.
            transform: Convert the final value to a different data type before
                returning it.

        Raises:
            TypeError: If ``get`` is provided but the value value is not a
                dictionary like object.

        Runs the following actions in order:

        1. :meth:`load` if ``load`` is provided.
        2. :meth:`runway.util.MutableMap.find` or :meth:`dict.get` depending
           on the data type if ``get`` is provided.
        3. Convert null value string to ``NoneType`` object. This includes string
           values of "None" and "null". This conversion is case insensitive.
        4. :meth:`transform` if ``transform`` is provided.

        """
        if load:
            value = cls.load(value, parser=load, **kwargs)
        if get:
            if isinstance(value, MutableMap):
                value = value.find(get)
            elif isinstance(value, dict):
                value = value.get(get)
            else:
                raise TypeError(
                    f'value must be dict type to use "get"; got type "{type(value)}"'
                )
        if (
            isinstance(value, str)
            and value.lower() in ["none", "null"]
            and transform != "str"  # retain string "as is" if should be str
        ):
            value = None
        if transform:
            return cls.transform(value, to_type=transform, **kwargs)
        if isinstance(value, MutableMap):
            LOGGER.debug("returning data from MutableMap")
            return value.data
        return value

    @classmethod
    def handle(
        cls,
        __value: str,
        context: Union[CfnginContext, RunwayContext],
        *__args: Any,
        provider: Optional[Provider] = None,
        **__kwargs: Any,
    ) -> Any:
        """Perform the lookup.

        Args:
            __value: Parameter(s) given to the lookup.
            context: The current context object.
            provider: CFNgin AWS provider.

        """
        raise NotImplementedError

    @classmethod
    def parse(cls, value: str) -> Tuple[str, Dict[str, str]]:
        """Parse the value passed to a lookup in a standardized way.

        Args:
            value: The raw value passed to a lookup.

        Returns:
            The lookup query and a dict of arguments

        """
        raw_value = read_value_from_path(value)

        colon_split = raw_value.split("::", 1)

        query = colon_split.pop(0)
        args: Dict[str, str] = cls._parse_args(colon_split[0]) if colon_split else {}

        return query, args

    @classmethod
    def _parse_args(cls, args: str) -> Dict[str, str]:
        """Convert a string into an args dict.

        Each arg should be seporated by  ``,``. The key and value should
        be seporated by ``=``. Any leading or following spaces are stripped.

        Args:
            args: A string containing arguments to be parsed. (e.g.
                ``'key1=value1, key2=value2'``)

        Returns:
            Dict of parsed args.

        """
        split_args = args.split(",")
        return {
            key.strip(): value.strip()
            for key, value in [arg.split("=", 1) for arg in split_args]
        }

    @classmethod
    def load(cls, value: Any, parser: Optional[str] = None, **kwargs: Any) -> Any:
        """Load a formatted string or object into a python data type.

        First action taken in :meth:`format_results`.
        If a lookup needs to handling loading data to process it before it
        enters :meth:`format_results`, is should use
        ``args.pop('load')`` to prevent the data from being loaded twice.

        Args:
            value: What is being loaded.
            parser: Name of the parser to use.

        Returns:
            The loaded value.

        """
        mapping = {
            "json": cls._load_json,
            "troposphere": cls._load_troposphere,
            "yaml": cls._load_yaml,
        }

        if not parser:
            return value

        return mapping[parser](value, **kwargs)

    @classmethod
    def _load_json(cls, value: Any, **_: Any) -> MutableMap:
        """Load a JSON string into a MutableMap.

        Args:
            value: JSON formatted string.

        """
        if not isinstance(value, str):
            raise TypeError(
                'value of type "%s" must of type "str" to use the "load=json" argument.'
            )
        result = json.loads(value)
        if isinstance(result, dict):
            return MutableMap(**result)
        return result

    @classmethod
    def _load_troposphere(cls, value: Any, **_: Any) -> MutableMap:
        """Load a Troposphere resource into a MutableMap.

        Args:
            value: Troposphere resource to contvert to a MutableMap for parsing.

        """
        if not isinstance(value, BaseAWSObject):
            raise TypeError(
                'value of type "%s" must of type "troposphere.'
                'BaseAWSObject" to use the "load=troposphere" option.'
            )
        if hasattr(value, "properties"):
            return MutableMap(**value.properties)
        raise NotImplementedError(
            '"load=troposphere" only supports BaseAWSObject with a "properties" object.'
        )

    @classmethod
    def _load_yaml(cls, value: Any, **_: Any) -> MutableMap:
        """Load a YAML string into a MutableMap.

        Args:
            value: YAML formatted string.

        """
        if not isinstance(value, str):
            raise TypeError(
                'value of type "%s" must of type "str" to use the "load=yaml" argument.'
            )
        result = yaml.safe_load(value)
        if isinstance(result, dict):
            return MutableMap(**result)
        return result

    @classmethod
    def transform(
        cls,
        value: Any,
        *,
        to_type: Optional[TransformToTypeLiteral] = "str",
        **kwargs: Any,
    ) -> Any:
        """Transform the result of a lookup into another datatype.

        Last action taken in :meth:`format_results`.
        If a lookup needs to handling transforming the data in a way that
        the base class can't support it should overwrite this method of the
        base class to register different transform methods.

        Args:
            value: What is to be transformed.
            to_type: The type the value will be transformed into.

        Returns:
            The transformed value.

        """
        mapping = {"bool": cls._transform_to_bool, "str": cls._transform_to_string}

        if not to_type:
            return value

        return mapping[to_type](value, **kwargs)  # type: ignore

    @classmethod
    def _transform_to_bool(cls, value: Any, **_: Any) -> bool:
        """Transform a string into a bool.

        Args:
            value: The value to be transformed into a bool.

        Raises:
            ValueError: The value provided was not a bool or string or
                the string could not be converted to a bool.

        """
        if isinstance(value, bool):
            return value
        if isinstance(value, str):
            return bool(str2bool(value))
        raise TypeError(
            f"Value must be a string or bool to use transform=bool. Got type {type(value)}."
        )

    @classmethod
    def _transform_to_string(
        cls, value: Any, *, delimiter: str = ",", indent: int = 0, **_: Any
    ) -> str:
        """Transform anything into a string.

        If the datatype of ``value`` is a list or similar to a list, ``join()``
        is used to construct the list using a given delimiter or ``,``.

        Args:
            value: The value to be transformed into a string.
            delimiter: Used when transforming a list like object into a string
                to join each element together.
            indent: Number of spaces to use when indenting JSON output.

        """
        if isinstance(value, (list, set, tuple)):
            return f"{delimiter}".join(cast(Sequence[str], value))
        if isinstance(value, MutableMap):
            # convert into a dict with protected attrs removed
            value = value.data
        if isinstance(value, dict):
            # dumped twice for an escaped json dict
            return json.dumps(
                json.dumps(cast(Dict[str, Any], value), indent=int(indent))
            )
        if isinstance(value, bool):
            return json.dumps(str(value))
        return str(value)



"""Retrieve the ID of the Cognito User Pool."""
import logging
from typing import Any, Dict, Optional

from ...base import HookArgsBaseModel

LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    created_user_pool_id: Optional[str] = None
    """The ID of the created Cognito User Pool."""

    user_pool_arn: Optional[str] = None
    """The ARN of the supplied User pool."""


def get(*__args: Any, **kwargs: Any) -> Dict[str, Any]:
    """Retrieve the ID of the Cognito User Pool.

    The User Pool can either be supplied via an ARN or by being generated.
    If the user has supplied an ARN that utilize that, otherwise retrieve
    the generated id. Used in multiple pre_hooks for Auth@Edge.

    Arguments parsed by
    :class:`~runway.cfngin.hooks.staticsite.auth_at_edge.user_pool_id_retriever.HookArgs`.

    """
    args = HookArgs.parse_obj(kwargs)

    # Favor a specific arn over a created one
    if args.user_pool_arn:
        return {"id": args.user_pool_arn.split("/")[-1:][0]}
    if args.created_user_pool_id:
        return {"id": args.created_user_pool_id}
    return {"id": ""}



"""AMI lookup."""
# pylint: disable=no-self-argument
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import operator
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from pydantic import validator
from typing_extensions import Final, Literal

from ....lookups.handlers.base import LookupHandler
from ....utils import BaseModel
from ...utils import read_value_from_path

if TYPE_CHECKING:
    from ....context import CfnginContext


class ArgsDataModel(BaseModel):
    """Arguments data model.

    Any other arguments specified are sent as filters to the AWS API.
    For example, ``architecture:x86_64`` will add a filter.

    """

    executable_users: Optional[List[str]] = None
    """List of executable users."""

    owners: List[str]
    """At least one owner is required.

    Should be ``amazon``, ``self``, or an AWS account ID.

    """

    region: Optional[str] = None
    """AWS region."""

    @validator("executable_users", "owners", allow_reuse=True, pre=True)
    def _convert_str_to_list(cls, v: Union[List[str], str]) -> List[str]:
        """Convert str to list."""
        if isinstance(v, str):
            return v.split(",")
        return v  # cov: ignore


class ImageNotFound(Exception):
    """Image not found."""

    search_string: str

    def __init__(self, search_string: str) -> None:
        """Instantiate class."""
        self.search_string = search_string
        super().__init__(
            f"Unable to find ec2 image with search string: {search_string}"
        )


class AmiLookup(LookupHandler):
    """AMI lookup."""

    TYPE_NAME: Final[Literal["ami"]] = "ami"
    """Name that the Lookup is registered as."""

    @classmethod
    def parse(cls, value: str) -> Tuple[str, Dict[str, str]]:
        """Parse the value passed to the lookup.

        This overrides the default parseing to account for special requirements.

        Args:
            value: The raw value passed to a lookup.

        Returns:
            The lookup query and a dict of arguments

        """
        raw_value = read_value_from_path(value)
        args: Dict[str, str] = {}

        if "@" in raw_value:
            args["region"], raw_value = raw_value.split("@", 1)

        # now find any other arguments that can be filters
        matches = re.findall(r"([0-9a-zA-z_-]+:[^\s$]+)", raw_value)
        for match in matches:
            k, v = match.split(":", 1)
            args[k] = v

        return args.pop("name_regex"), args

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls, value: str, context: CfnginContext, *__args: Any, **__kwargs: Any
    ) -> str:
        """Fetch the most recent AMI Id using a filter.

        Args:
            value: Parameter(s) given to this lookup.
            context: Context instance.

        Example:
            .. code-block:

                ${ami [<region>@]owners:self,account,amazon name_regex:serverX-[0-9]+ architecture:x64,i386}

            The above fetches the most recent AMI where owner is self
            account or amazon and the ami name matches the regex described,
            the architecture will be either x64 or i386

            You can also optionally specify the region in which to perform the
            AMI lookup.

        """  # noqa
        query, raw_args = cls.parse(value)
        args = ArgsDataModel.parse_obj(raw_args)
        ec2 = context.get_session(region=args.region).client("ec2")

        describe_args: Dict[str, Any] = {
            "Filters": [
                {"Name": key, "Values": val.split(",") if val else val}
                for key, val in {
                    k: v
                    for k, v in raw_args.items()
                    if k not in ArgsDataModel.__fields__
                }.items()
            ],
            "Owners": args.owners,
        }
        if args.executable_users:
            describe_args["ExecutableUsers"] = args.executable_users

        result = ec2.describe_images(**describe_args)

        images = sorted(
            result.get("Images", []),
            key=operator.itemgetter("CreationDate"),
            reverse=True,
        )
        for image in images:
            # sometimes we get ARI/AKI in response - these don't have a 'Name'
            if re.match(f"^{query}$", image.get("Name", "")) and "ImageId" in image:
                return image["ImageId"]

        raise ImageNotFound(value)



"""File generator.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/filegenerator.py

"""
from __future__ import annotations

import datetime
import os
import stat
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from queue import Queue
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from botocore.exceptions import ClientError
from dateutil.parser import parse
from dateutil.tz import tzlocal
from typing_extensions import TypedDict

from .utils import (
    EPOCH_TIME,
    BucketLister,
    create_warning,
    find_bucket_key,
    find_dest_path_comp_key,
    get_file_stat,
)

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import HeadObjectOutputTypeDef, ObjectTypeDef

    from ......type_defs import AnyPath
    from .format_path import FormatPathResult, SupportedPathType


def is_readable(path: Path) -> bool:
    """Check to see if a file or a directory can be read.

    This is tested by performing an operation that requires read access
    on the file or the directory.

    """
    if path.is_dir():
        try:
            os.listdir(path)
        except OSError:
            return False
    else:
        try:
            with open(path, "r", encoding="utf-8"):
                pass
        except OSError:
            return False
    return True


def is_special_file(path: Path) -> bool:
    """Check to see if a special file.

    It checks if the file is a character special device, block special device,
    FIFO, or socket.

    """
    mode = path.stat().st_mode
    # Character special device.
    if stat.S_ISCHR(mode):
        return True
    # Block special device
    if stat.S_ISBLK(mode):
        return True
    # FIFO.
    if stat.S_ISFIFO(mode):
        return True
    # Socket.
    if stat.S_ISSOCK(mode):
        return True
    return False


FileStatsDict = TypedDict(
    "FileStatsDict",
    src="AnyPath",
    compare_key=Optional[str],
    dest_type=Optional["SupportedPathType"],
    dest=Optional[str],
    last_update=datetime.datetime,
    operation_name=Optional[str],
    response_data=Optional[Union["HeadObjectOutputTypeDef", "ObjectTypeDef"]],
    size=Optional[int],
    src_type=Optional["SupportedPathType"],
)


@dataclass
class FileStats:
    """Information about a file.

    Attributes:
        src: Path to source file or S3 object.
        compare_key: The name of the file relative to the specified
                directory/prefix. This variable is used when performing synching
                or if the destination file is adopting the source file's name.
        dest_type: Location of the destination - either local or S3.
        dest: Path to destination file or S3 object.
        last_update: Timestamp when file or S3 object was last updated.
        operation_name: Name of the operation being run.
        response_data: S3 object metadata. Only set if dest_type == S3.
        size: Size of the file or S3 object in bytes.
        src_type: Location of the source - either local or S3.

    """

    src: AnyPath
    compare_key: Optional[str] = None
    dest: Optional[str] = None
    dest_type: Optional[SupportedPathType] = None
    last_update: datetime.datetime = EPOCH_TIME
    operation_name: Optional[str] = None
    response_data: Optional[Union[HeadObjectOutputTypeDef, ObjectTypeDef]] = None
    size: Optional[int] = None
    src_type: Optional[SupportedPathType] = None

    def dict(self) -> FileStatsDict:
        """Dump contents of object to a dict."""
        return deepcopy(cast(FileStatsDict, self.__dict__))


_LastModifiedAndSize = TypedDict(
    "_LastModifiedAndSize", Size=int, LastModified=datetime.datetime
)


class FileGenerator:
    """Create a generator to yield files.

    It is universal in the sense that it will handle s3 files, local files,
    local directories, and s3 objects under the same common prefix.

    """

    result_queue: "Queue[Any]"

    def __init__(
        self,
        client: S3Client,
        operation_name: str,
        follow_symlinks: bool = True,
        page_size: Optional[int] = None,
        result_queue: Optional["Queue[Any]"] = None,
        request_parameters: Any = None,
    ):
        """Instantiate class.

        Args:
            client: boto3 S3 client.
            operation_name: Name of the operation being executed.
            follow_symlinks: If symlinks should be followed.
            page_size: Number of items per page.
            result_queue: Queue used for outputing results.
            request_parameters: Parameters provided with the request.

        """
        self._client = client
        self.operation_name = operation_name
        self.follow_symlinks = follow_symlinks
        self.page_size = page_size
        self.result_queue = result_queue or Queue()
        self.request_parameters = {}
        if request_parameters is not None:
            self.request_parameters = request_parameters

    def call(self, files: FormatPathResult) -> Generator[FileStats, None, None]:
        """Generalized function to yield the ``FileInfo`` objects."""
        function_table = {"s3": self.list_objects, "local": self.list_files}
        file_iterator = function_table[files["src"]["type"]](
            files["src"]["path"], files["dir_op"]
        )
        for src_path, extra_information in file_iterator:
            dest_path, compare_key = find_dest_path_comp_key(files, src_path)
            file_stat_kwargs: FileStatsDict = {
                "compare_key": compare_key,
                "dest": dest_path,
                "dest_type": files["dest"]["type"],
                "last_update": extra_information.get("LastModified", EPOCH_TIME),
                "operation_name": self.operation_name,
                "response_data": None,
                "size": extra_information.get("Size", 0),
                "src": src_path,
                "src_type": files["src"]["type"],
            }
            if files["src"]["type"] == "s3":
                file_stat_kwargs["response_data"] = cast(
                    Optional[Union["HeadObjectOutputTypeDef", "ObjectTypeDef"]],
                    extra_information,
                )
            yield FileStats(**file_stat_kwargs)

    def list_files(
        self, path: AnyPath, dir_op: bool
    ) -> Generator[Tuple[Path, _LastModifiedAndSize], None, None]:
        """Yield the appropriate local file or local files under a directory.

        For directories a depth first search is implemented in order to
        follow the same sorted pattern as a s3 list objects operation
        outputs. It yields the file's source path, size, and last
        update.

        """
        if isinstance(path, str):
            path = Path(path)
        if not self.should_ignore_file(path):
            if not dir_op:
                stats = self.safely_get_file_stats(path)
                if stats:
                    yield stats
            else:
                # using os.listdir instead of Path.iterdir so we can sort the list
                # but not load the entire tree into memory
                listdir_names = os.listdir(path)
                names: List[str] = []
                for name in listdir_names:
                    if (path / name).is_dir():
                        name = name + os.path.sep
                    names.append(name)
                self.normalize_sort(names, os.sep, "/")
                for name in names:
                    file_path = path / name
                    if file_path.is_dir():
                        for result in self.list_files(file_path, dir_op):
                            yield result
                    else:
                        stats = self.safely_get_file_stats(file_path)
                        if stats:
                            yield stats

    @staticmethod
    def normalize_sort(names: List[str], os_sep: str, character: str) -> None:
        """Ensure that the same path seperator is used when sorting.

        On Windows, the path operator is a backslash as opposed to a forward slash
        which can lead to differences in sorting between S3 and a Windows machine.

        Args:
            names: List of file names.
            os_sep: OS seperator.
            character: Character that will be used to replace the os_sep.

        """
        names.sort(key=lambda item: item.replace(os_sep, character))

    def safely_get_file_stats(
        self, path: Path
    ) -> Optional[Tuple[Path, _LastModifiedAndSize]]:
        """Get file stats with handling for some common errors.

        Args:
            path: Path to a file.

        """
        try:
            size, last_update = get_file_stat(path)
        except (OSError, ValueError):
            self.triggers_warning(path)
        else:
            last_update = self._validate_update_time(last_update, path)
            return path, {"Size": size, "LastModified": last_update}
        return None

    def _validate_update_time(
        self, update_time: Optional[datetime.datetime], path: Path
    ) -> datetime.datetime:
        """Handle missing last modified time."""
        if update_time is None:
            warning = create_warning(
                path=path,
                error_message="File has an invalid timestamp. Passing epoch "
                "time as timestamp.",
                skip_file=False,
            )
            self.result_queue.put(warning)
            return EPOCH_TIME
        return update_time

    def should_ignore_file(self, path: Path) -> bool:
        """Check whether a file should be ignored in the file generation process.

        This includes symlinks that are not to be followed and files that generate
        warnings.

        """
        if not self.follow_symlinks:
            if path.is_dir() and path.is_symlink():
                # is_symlink returns False if it does not exist
                return True
        warning_triggered = self.triggers_warning(path)
        if warning_triggered:
            return True
        return False

    def triggers_warning(self, path: Path) -> bool:
        """Check the specific types and properties of a file.

        If the file would cause trouble, the function adds a
        warning to the result queue to be printed out and returns a boolean
        value notify whether the file caused a warning to be generated.
        Files that generate warnings are skipped. Currently, this function
        checks for files that do not exist and files that the user does
        not have read access.

        """
        if not path.exists():
            warning = create_warning(path, "File does not exist.")
            self.result_queue.put(warning)
            return True
        if is_special_file(path):
            warning = create_warning(
                path,
                (
                    "File is character special device, "
                    "block special device, FIFO, or socket."
                ),
            )
            self.result_queue.put(warning)
            return True
        if not is_readable(path):
            warning = create_warning(path, "File/Directory is not readable.")
            self.result_queue.put(warning)
            return True
        return False

    def list_objects(
        self, s3_path: str, dir_op: bool
    ) -> Generator[
        Tuple[str, Union[HeadObjectOutputTypeDef, ObjectTypeDef]], None, None
    ]:
        """Yield the appropriate object or objects under a common prefix.

        It yields the file's source path, size, and last update.

        Args:
            s3_path: Path to list.
            dir_op: If the path is a directory.

        """
        # Short circuit path: if we are not recursing into the s3
        # bucket and a specific path was given, we can just yield
        # that path and not have to call any operation in s3.
        bucket, prefix = find_bucket_key(s3_path)
        if not dir_op and prefix:
            yield self._list_single_object(s3_path)
        else:
            lister = BucketLister(self._client)
            extra_args: Any = self.request_parameters.get("ListObjectsV2", {})
            for obj in lister.list_objects(
                bucket=bucket,
                prefix=prefix,
                page_size=self.page_size,
                extra_args=extra_args,
            ):
                source_path, response_data = obj
                if response_data.get("Size", 0) == 0 and source_path.endswith("/"):
                    if self.operation_name == "delete":
                        # This is to filter out manually created folders
                        # in S3.  They have a size zero and would be
                        # undesirably downloaded.  Local directories
                        # are automatically created when they do not
                        # exist locally.  But user should be able to
                        # delete them.
                        yield source_path, response_data
                elif not dir_op and s3_path != source_path:
                    pass
                else:
                    yield source_path, response_data

    def _list_single_object(self, s3_path: str) -> Tuple[str, HeadObjectOutputTypeDef]:
        """List single object."""
        # When we know we're dealing with a single object, we can avoid
        # a ListObjects operation (which causes concern for anyone setting
        # IAM policies with the smallest set of permissions needed) and
        # instead use a HeadObject request.
        if self.operation_name == "delete":
            # If the operation is just a single remote delete, there is
            # no need to run HeadObject on the S3 object as none of the
            # information gained from HeadObject is required to delete the
            # object.
            return s3_path, {"Size": None, "LastModified": None}  # type: ignore
        bucket, key = find_bucket_key(s3_path)
        try:
            params: Dict[str, Any] = {"Bucket": bucket, "Key": key}
            # params.update(self.request_parameters.get("HeadObject", {}))
            response = self._client.head_object(**params)
        except ClientError as exc:
            # We want to try to give a more helpful error message.
            # This is what the customer is going to see so we want to
            # give as much detail as we have.
            if not exc.response["Error"]["Code"] == "404":
                raise
            # The key does not exist so we'll raise a more specific
            # error message here.
            response = exc.response.copy()
            response["Error"]["Message"] = f'Key "{key}" does not exist'
            raise ClientError(response, "HeadObject") from None
        response["Size"] = int(response.pop("ContentLength"))  # type: ignore
        last_update = parse(response["LastModified"])  # type: ignore
        response["LastModified"] = last_update.astimezone(tzlocal())
        return s3_path, response



"""Retrieve a value from CloudFormation Stack Outputs.

The query syntax for this lookup is ``<stack-name>.<output-name>``.
When specifying the output name, be sure to use the *Logical ID* of
the output; not the *Export.Name*.

"""
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Any, Dict, NamedTuple, Optional, Union, cast

from botocore.exceptions import ClientError
from typing_extensions import Final, Literal

from ...cfngin.exceptions import StackDoesNotExist
from ...exceptions import OutputDoesNotExist
from .base import LookupHandler

if TYPE_CHECKING:
    from mypy_boto3_cloudformation.client import CloudFormationClient

    from ...cfngin.providers.aws.default import Provider
    from ...context import CfnginContext, RunwayContext

LOGGER = logging.getLogger(__name__)


class OutputQuery(NamedTuple):
    """Output query NamedTuple."""

    stack_name: str
    output_name: str


class CfnLookup(LookupHandler):
    """CloudFormation Stack Output lookup."""

    TYPE_NAME: Final[Literal["cfn"]] = "cfn"
    """Name that the Lookup is registered as."""

    @staticmethod
    def should_use_provider(args: Dict[str, str], provider: Optional[Provider]) -> bool:
        """Determine if the provider should be used for the lookup.

        This will open happen when the lookup is used with CFNgin.

        Args:
            args: Parsed arguments provided to the lookup.
            provider: CFNgin provider.

        """
        if provider:
            if args.get("region") and provider.region != args["region"]:
                LOGGER.debug("not using provider; requested region does not match")
                return False
            LOGGER.debug("using provider")
            return True
        return False

    @staticmethod
    def get_stack_output(client: CloudFormationClient, query: OutputQuery) -> str:
        """Get CloudFormation Stack output.

        Args:
            client: Boto3 CloudFormation client.
            query: What to get.

        """
        LOGGER.debug("describing stack: %s", query.stack_name)
        stack = client.describe_stacks(StackName=query.stack_name)["Stacks"][0]
        outputs = {
            # these should always exist even though the schema says they are not required
            output["OutputKey"]: output["OutputValue"]  # type: ignore
            for output in stack.get("Outputs", [])
        }
        LOGGER.debug("%s stack outputs: %s", stack["StackName"], json.dumps(outputs))
        return outputs[query.output_name]

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls,
        value: str,
        context: Union[CfnginContext, RunwayContext],
        *,
        provider: Optional[Provider] = None,
        **_: Any,
    ) -> Any:
        """Retrieve a value from CloudFormation Stack outputs.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.
            provider: AWS provider.

        Returns:
            Result of the query.

        Raises:
            OutputDoesNotExist: Output does not exist on the Stack provided
                and default was not provided.

        """
        raw_query, args = cls.parse(value)
        try:
            query = OutputQuery(*raw_query.split("."))
        except TypeError as exc:
            raise ValueError(
                f'query must be <stack-name>.<output-name>; got "{raw_query}"'
            ) from exc

        try:
            # dict is not perserved in mock call so it must be a copy of
            # args for testing to function correctly
            if cls.should_use_provider(args.copy(), provider):
                # this will only happen when used from cfngin
                result = cast("Provider", provider).get_output(
                    query.stack_name, query.output_name
                )
            else:
                cfn_client = context.get_session(region=args.get("region")).client(
                    "cloudformation"
                )
                result = cls.get_stack_output(cfn_client, query)
        except (ClientError, KeyError, StackDoesNotExist) as exc:
            # StackDoesNotExist is only raised by provider
            if "default" in args:
                LOGGER.debug(
                    "unable to resolve lookup for CloudFormation Stack "
                    'output "%s"; using default',
                    raw_query,
                    exc_info=True,
                )
                args.pop("load", None)  # don't load a default value
                result = args.pop("default")
            elif isinstance(exc, (ClientError, StackDoesNotExist)):
                raise
            else:
                raise OutputDoesNotExist(query.stack_name, query.output_name) from exc
        return cls.format_results(result, **args)



"""AWS ECS hook."""
# pylint: disable=no-self-argument
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Union

from pydantic import validator
from typing_extensions import TypedDict

from ...utils import BaseModel

if TYPE_CHECKING:
    from mypy_boto3_ecs.type_defs import CreateClusterResponseTypeDef

    from ...context import CfnginContext

LOGGER = logging.getLogger(__name__)


class CreateClustersHookArgs(BaseModel):
    """Hook arguments for ``create_clusters``."""

    clusters: List[str]
    """List of cluster names to create."""

    @validator("clusters", allow_reuse=True, pre=True)
    def _convert_clusters(cls, v: Union[List[str], str]) -> List[str]:
        """Convert value of ``clusters`` from str to list."""
        if isinstance(v, str):
            return [v]
        return v


class CreateClustersResponseTypeDef(TypedDict):
    """Response from create_clusters."""

    clusters: Dict[str, CreateClusterResponseTypeDef]


def create_clusters(
    context: CfnginContext, *__args: Any, **kwargs: Any
) -> CreateClustersResponseTypeDef:
    """Create ECS clusters.

    Args:
        context: CFNgin context object.

    """
    args = CreateClustersHookArgs.parse_obj(kwargs)
    ecs_client = context.get_session().client("ecs")

    cluster_info: Dict[str, Any] = {}
    for cluster in args.clusters:
        LOGGER.debug("creating ECS cluster: %s", cluster)
        response = ecs_client.create_cluster(clusterName=cluster)
        cluster_info[response.get("cluster", {}).get("clusterName", "")] = response
    return {"clusters": cluster_info}



"""Generate a random string."""
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import logging
import secrets
import string
from typing import TYPE_CHECKING, Any, Callable, List, Sequence, Union

from typing_extensions import Final, Literal

from ...utils import BaseModel
from .base import LookupHandler

if TYPE_CHECKING:
    from ...context import CfnginContext, RunwayContext

LOGGER = logging.getLogger(__name__)


class ArgsDataModel(BaseModel):
    """Arguments data model."""

    digits: bool = True
    lowercase: bool = True
    punctuation: bool = False
    uppercase: bool = True


class RandomStringLookup(LookupHandler):
    """Random string lookup."""

    TYPE_NAME: Final[Literal["random.string"]] = "random.string"
    """Name that the Lookup is registered as."""

    @staticmethod
    def calculate_char_set(args: ArgsDataModel) -> str:
        """Calculate character set from the provided arguments."""
        char_set = ""
        if args.digits:
            char_set += string.digits
        if args.lowercase:
            char_set += string.ascii_lowercase
        if args.punctuation:
            char_set += string.punctuation
        if args.uppercase:
            char_set += string.ascii_uppercase
        LOGGER.debug("character set: %s", char_set)
        return char_set

    @staticmethod
    def generate_random_string(char_set: Sequence[str], length: int) -> str:
        """Generate a random string of a set length from a set of characters."""
        return "".join(secrets.choice(char_set) for _ in range(length))

    @staticmethod
    def has_digit(value: str) -> bool:
        """Check if value contains a digit."""
        return any(v.isdigit() for v in value)

    @staticmethod
    def has_lowercase(value: str) -> bool:
        """Check if value contains lowercase."""
        return any(v.islower() for v in value)

    @staticmethod
    def has_punctuation(value: str) -> bool:
        """Check if value contains uppercase."""
        return any(v in string.punctuation for v in value)

    @staticmethod
    def has_uppercase(value: str) -> bool:
        """Check if value contains uppercase."""
        return any(v.isupper() for v in value)

    @classmethod
    def ensure_has_one_of(cls, args: ArgsDataModel, value: str) -> bool:
        """Ensure value has at least one of each required character.

        Args:
            args: Hook args.
            value: Value to check.

        """
        checks: List[Callable[[str], bool]] = []
        if args.digits:
            checks.append(cls.has_digit)
        if args.lowercase:
            checks.append(cls.has_lowercase)
        if args.punctuation:
            checks.append(cls.has_punctuation)
        if args.uppercase:
            checks.append(cls.has_uppercase)
        return sum(c(value) for c in checks) == len(checks)

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls,
        value: str,
        context: Union[CfnginContext, RunwayContext],
        *__args: Any,
        **__kwargs: Any,
    ) -> Any:
        """Generate a random string.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ValueError: Unable to find a value for the provided query and
                a default value was not provided.

        """
        raw_length, raw_args = cls.parse(value)
        length = int(raw_length)
        args = ArgsDataModel.parse_obj(raw_args)
        char_set = cls.calculate_char_set(args)
        while True:
            result = cls.generate_random_string(char_set, length)
            if cls.ensure_has_one_of(args, result):
                break
        return cls.format_results(result, **raw_args)



"""Check the authorization information passed along via the cookie headers.

When the information is not present or an error occurs due to verification a new
request to Cognito will be made to authorize the user. The user will be taken
to the Cognito login page to enter their information, and as long as it is valid,
the information will be passed to the parsing agent.

If a refresh token exists and it has expired (after 1 hour) do an automatic refresh
of the credentials by redirecting the user to the refresh agent.

"""
# pylint: disable=consider-using-f-string
import base64
import datetime
import hashlib
import json
import logging
import re
import secrets
from urllib.parse import quote_plus, urlencode

from shared_jose import validate_jwt  # noqa pylint: disable=import-error

from shared import (  # noqa pylint: disable=import-error
    decode_token,
    extract_and_parse_cookies,
    get_config,
    sign,
    timestamp_in_seconds,
)

LOGGER = logging.getLogger(__file__)

SECRET_ALLOWED_CHARS = (
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~"
)
NONCE_LENGTH = 16
PKCE_LENGTH = 43
CONFIG = get_config()


def handler(event, _context):
    """Handle the request passed in.

    Args:
        event (Dict[str, Any]): The Lambda Event.
        _context (Any): Lambda context object.

    """
    request = event["Records"][0]["cf"]["request"]
    domain_name = request["headers"]["host"][0]["value"]
    querystring = request.get("querystring")
    request_query_string = f"?{querystring}" if querystring else ""
    requested_uri = f"{request['uri']}{request_query_string}"

    try:
        # Extract the cookies received from Cognito
        extracted = extract_and_parse_cookies(request["headers"], CONFIG["client_id"])
        token_user_name = extracted["token_user_name"]
        id_token = extracted["id_token"]
        refresh_token = extracted["refresh_token"]

        # If the token user name or id token are missing then we need
        # new credentials
        if not token_user_name or not id_token:
            msg = "No valid credentials present in cookies"
            LOGGER.error(msg)
            raise Exception(msg)

        # Get the expiration date from the id token.
        decoded_token = decode_token(id_token)
        expiration = decoded_token.get("exp")
        exp_date = datetime.datetime.fromtimestamp(expiration)
        now = datetime.datetime.now()

        # If we have a refresh token and the expiration date has passed
        # then forward the user to the refresh agent
        if now > (exp_date - datetime.timedelta(minutes=10)) and refresh_token:
            nonce = generate_nonce()
            response = {
                "status": "307",
                "statusDescription": "Temporary Redirect",
                "headers": {
                    # Redirect the user to the refresh agent
                    "location": [
                        {
                            "key": "location",
                            "value": "https://%s%s?%s"
                            % (
                                domain_name,
                                CONFIG.get("redirect_path_auth_refresh"),
                                urlencode(
                                    {"requestedUri": requested_uri, "nonce": nonce}
                                ),
                            ),
                        }
                    ],
                    "set-cookie": [
                        # Set the Nonce Cookie
                        {
                            "key": "set-cookie",
                            "value": "spa-auth-edge-nonce=%s; %s"
                            % (
                                quote_plus(nonce),
                                CONFIG.get("cookie_settings").get("nonce"),
                            ),
                        },
                        {
                            "key": "set-cookie",
                            "value": "spa-auth-edge-nonce-hmac=%s; %s"
                            % (
                                quote_plus(sign(nonce, CONFIG["nonce_signing_secret"])),
                                CONFIG.get("cookie_settings").get("nonce"),
                            ),
                        },
                    ],
                    **CONFIG.get("cloud_front_headers", {}),
                },
            }
            return response

        # Validate the token information against the Cognito JWKS
        validate_jwt(
            id_token,
            CONFIG["token_jwks_uri"],
            CONFIG["token_issuer"],
            CONFIG["client_id"],
        )

        return request
    except Exception:  # noqa pylint: disable=broad-except
        # We need new authorization. Get the user over to Cognito
        nonce = generate_nonce()
        state = {
            "nonce": nonce,
            "nonceHmac": sign(nonce, CONFIG["nonce_signing_secret"]),
            **generate_pkce_verifier(),
        }
        login_query_string = urlencode(
            {
                "redirect_uri": "https://%s%s"
                % (domain_name, CONFIG["redirect_path_sign_in"]),
                "response_type": "code",
                "client_id": CONFIG["client_id"],
                "state": base64.urlsafe_b64encode(
                    bytes(
                        json.dumps(
                            {"nonce": state["nonce"], "requestedUri": requested_uri}
                        ).encode()
                    )
                ),
                "scope": " ".join(CONFIG["oauth_scopes"]),
                "code_challenge_method": "S256",
                "code_challenge": state["pkceHash"],
            },
            quote_via=quote_plus,
        )
        # Redirect user to the Cognito Login
        response = {
            "status": "307",
            "statusDescription": "Temporary Redirect",
            "headers": {
                "location": [
                    {
                        "key": "location",
                        "value": "https://%s/oauth2/authorize?%s"
                        % (CONFIG["cognito_auth_domain"], login_query_string),
                    }
                ],
                "set-cookie": [
                    {
                        "key": "set-cookie",
                        "value": "spa-auth-edge-nonce=%s; %s"
                        % (
                            quote_plus(state["nonce"]),
                            CONFIG.get("cookie_settings", {}).get("nonce"),
                        ),
                    },
                    {
                        "key": "set-cookie",
                        "value": "spa-auth-edge-nonce-hmac=%s; %s"
                        % (
                            quote_plus(state["nonceHmac"]),
                            CONFIG.get("cookie_settings", {}).get("nonce"),
                        ),
                    },
                    {
                        "key": "set-cookie",
                        "value": "spa-auth-edge-pkce=%s; %s"
                        % (
                            quote_plus(state["pkce"]),
                            CONFIG.get("cookie_settings", {}).get("nonce"),
                        ),
                    },
                ],
                **CONFIG.get("cloud_front_headers", {}),
            },
        }
        return response


def generate_pkce_verifier():
    """Generate the PKCE verification code."""
    pkce = random_key(PKCE_LENGTH)
    pkce_hash = hashlib.sha256()
    pkce_hash.update(pkce.encode("UTF-8"))
    pkce_hash = pkce_hash.digest()
    pkce_hash_b64 = base64.b64encode(pkce_hash)
    decoded_pkce = pkce_hash_b64.decode("UTF-8")
    decoded_pkce = re.sub(r"\=", "", decoded_pkce)
    decoded_pkce = re.sub(r"\+", "-", decoded_pkce)
    decoded_pkce = re.sub(r"\/", "_", decoded_pkce)

    return {"pkce": pkce, "pkceHash": decoded_pkce}


def generate_nonce():
    """Generate a random Nonce token."""
    return str(timestamp_in_seconds()) + "T" + random_key(NONCE_LENGTH)


def random_key(length=15):
    """Generate a random key of specified length from the allowed secret characters.

    Args:
        length (int): The length of the random key.

    """
    return "".join(secrets.choice(SECRET_ALLOWED_CHARS) for _ in range(length))



"""Runway Serverless Framework Module options."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import Extra

from ...base import ConfigProperty


class RunwayServerlessPromotezipOptionDataModel(ConfigProperty):
    """Model for Runway Serverless module promotezip option."""

    bucketname: Optional[str] = None

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway Serverless Framework Module promotezip option"

    def __bool__(self) -> bool:
        """Evaluate the boolean value of the object instance."""
        return bool(self.dict(exclude_none=True))


class RunwayServerlessModuleOptionsDataModel(ConfigProperty):
    """Model for Runway Serverless Framework Module options."""

    args: List[str] = []
    extend_serverless_yml: Dict[str, Any] = {}
    promotezip: RunwayServerlessPromotezipOptionDataModel = (
        RunwayServerlessPromotezipOptionDataModel()
    )
    skip_npm_ci: bool = False

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.ignore
        title = "Runway Serverless Framework Module options"



"""Botocore with support for AWS SSO credential assets."""
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import json
import logging
import os
from hashlib import sha1

from botocore import UNSIGNED
from botocore.config import Config
from botocore.credentials import (
    AssumeRoleProvider,
    BotoProvider,
    CachedCredentialFetcher,
    CanonicalNameCredentialSourcer,
    ContainerProvider,
    CredentialProvider,
    CredentialResolver,
    DeferredRefreshableCredentials,
    EnvProvider,
    InstanceMetadataFetcher,
    InstanceMetadataProvider,
    JSONFileCache,
    OriginalEC2Provider,
)
from botocore.credentials import (
    ProfileProviderBuilder as BotocoreProfileProviderBuilder,
)
from botocore.credentials import _get_client_creator, _serialize_if_needed
from botocore.exceptions import InvalidConfigError
from dateutil.tz import tzutc

from .exceptions import UnauthorizedSSOTokenError
from .util import SSOTokenLoader

LOGGER = logging.getLogger(__name__)


def create_credential_resolver(session, cache=None, region_name=None):
    """Create a default credential resolver.

    This creates a pre-configured credential resolver
    that includes the default lookup chain for
    credentials.

    """
    profile_name = session.get_config_variable("profile") or "default"
    metadata_timeout = session.get_config_variable("metadata_service_timeout")
    num_attempts = session.get_config_variable("metadata_service_num_attempts")
    disable_env_vars = session.instance_variables().get("profile") is not None

    if cache is None:
        cache = {}

    env_provider = EnvProvider()
    container_provider = ContainerProvider()
    instance_metadata_provider = InstanceMetadataProvider(
        iam_role_fetcher=InstanceMetadataFetcher(
            timeout=metadata_timeout,
            num_attempts=num_attempts,
            user_agent=session.user_agent(),
        )
    )

    profile_provider_builder = ProfileProviderBuilder(
        session, cache=cache, region_name=region_name
    )
    assume_role_provider = AssumeRoleProvider(
        load_config=lambda: session.full_config,
        client_creator=_get_client_creator(session, region_name),
        cache=cache,
        profile_name=profile_name,
        credential_sourcer=CanonicalNameCredentialSourcer(
            [env_provider, container_provider, instance_metadata_provider]
        ),
        profile_provider_builder=profile_provider_builder,
    )

    pre_profile = [
        env_provider,
        assume_role_provider,
    ]
    profile_providers = profile_provider_builder.providers(
        profile_name=profile_name, disable_env_vars=disable_env_vars,
    )
    post_profile = [
        OriginalEC2Provider(),
        BotoProvider(),
        container_provider,
        instance_metadata_provider,
    ]
    providers = pre_profile + profile_providers + post_profile

    if disable_env_vars:
        # An explicitly provided profile will negate an EnvProvider.
        # We will defer to providers that understand the "profile"
        # concept to retrieve credentials.
        # The one edge case if is all three values are provided via
        # env vars:
        # export AWS_ACCESS_KEY_ID=foo
        # export AWS_SECRET_ACCESS_KEY=bar
        # export AWS_PROFILE=baz
        # Then, just like our client() calls, the explicit credentials
        # will take precedence.
        #
        # This precedence is enforced by leaving the EnvProvider in the chain.
        # This means that the only way a "profile" would win is if the
        # EnvProvider does not return credentials, which is what we want
        # in this scenario.
        providers.remove(env_provider)
        LOGGER.debug(
            "Skipping environment variable credential check"
            " because profile name was explicitly set."
        )

    return CredentialResolver(providers=providers)


class ProfileProviderBuilder(BotocoreProfileProviderBuilder):
    """Extends the botocore profile provider builder to support AWS SSO."""

    def __init__(self, session, cache=None, region_name=None, sso_token_cache=None):
        """Instantiate class."""
        super().__init__(session, cache, region_name)
        self._sso_token_cache = sso_token_cache

    def providers(self, profile_name, disable_env_vars=False):
        """Return list of providers."""
        return [
            self._create_web_identity_provider(profile_name, disable_env_vars,),
            self._create_sso_provider(profile_name),
            self._create_shared_credential_provider(profile_name),
            self._create_process_provider(profile_name),
            self._create_config_provider(profile_name),
        ]

    def _create_sso_provider(self, profile_name):
        """AWS SSO credential provider."""
        return SSOProvider(
            load_config=lambda: self._session.full_config,
            client_creator=self._session.create_client,
            profile_name=profile_name,
            cache=self._cache,
            token_cache=self._sso_token_cache,
        )


class SSOCredentialFetcher(CachedCredentialFetcher):
    """AWS SSO credential fetcher."""

    def __init__(
        self,
        start_url,
        sso_region,
        role_name,
        account_id,
        client_creator,
        token_loader=None,
        cache=None,
        expiry_window_seconds=None,
    ):
        """Instantiate class."""
        self._client_creator = client_creator
        self._sso_region = sso_region
        self._role_name = role_name
        self._account_id = account_id
        self._start_url = start_url
        self._token_loader = token_loader

        super().__init__(cache, expiry_window_seconds)

    def _create_cache_key(self):
        """Create a predictable cache key for the current configuration.

        The cache key is intended to be compatible with file names.

        """
        args = {
            "startUrl": self._start_url,
            "roleName": self._role_name,
            "accountId": self._account_id,
        }
        # NOTE: It would be good to hoist this cache key construction logic
        # into the CachedCredentialFetcher class as we should be consistent.
        # Unfortunately, the current assume role fetchers that sub class don't
        # pass separators resulting in non-minified JSON. In the long term,
        # all fetchers should use the below caching scheme.
        args = json.dumps(args, sort_keys=True, separators=(",", ":"))
        argument_hash = sha1(args.encode("utf-8")).hexdigest()
        return self._make_file_safe(argument_hash)

    def _parse_timestamp(self, timestamp_ms):
        """Parse timestamp."""
        # fromtimestamp expects seconds so: milliseconds / 1000 = seconds
        timestamp_seconds = timestamp_ms / 1000.0
        timestamp = datetime.datetime.fromtimestamp(timestamp_seconds, tzutc())
        return _serialize_if_needed(timestamp)

    def _get_credentials(self):
        """Get credentials by calling SSO get role credentials."""
        config = Config(signature_version=UNSIGNED, region_name=self._sso_region,)
        client = self._client_creator("sso", config=config)

        kwargs = {
            "roleName": self._role_name,
            "accountId": self._account_id,
            "accessToken": self._token_loader(self._start_url),
        }
        try:
            response = client.get_role_credentials(**kwargs)
        except client.exceptions.UnauthorizedException as exc:
            raise UnauthorizedSSOTokenError() from exc
        credentials = response["roleCredentials"]

        credentials = {
            "ProviderType": "sso",
            "Credentials": {
                "AccessKeyId": credentials["accessKeyId"],
                "SecretAccessKey": credentials["secretAccessKey"],
                "SessionToken": credentials["sessionToken"],
                "Expiration": self._parse_timestamp(credentials["expiration"]),
            },
        }
        return credentials


class SSOProvider(CredentialProvider):
    """AWS SSO credential provider."""

    METHOD = "sso"

    _SSO_TOKEN_CACHE_DIR = os.path.expanduser(os.path.join("~", ".aws", "sso", "cache"))
    _SSO_CONFIG_VARS = [
        "sso_start_url",
        "sso_region",
        "sso_role_name",
        "sso_account_id",
    ]

    # pylint: disable=super-init-not-called
    def __init__(
        self, load_config, client_creator, profile_name, cache=None, token_cache=None
    ):
        """Instantiate class."""
        if token_cache is None:
            token_cache = JSONFileCache(self._SSO_TOKEN_CACHE_DIR)
        self._token_cache = token_cache
        if cache is None:
            cache = {}
        self.cache = cache
        self._load_config = load_config
        self._client_creator = client_creator
        self._profile_name = profile_name

    def _load_sso_config(self):
        """Load sso config."""
        loaded_config = self._load_config()
        profiles = loaded_config.get("profiles", {})
        profile_name = self._profile_name
        profile_config = profiles.get(self._profile_name, {})

        if all(c not in profile_config for c in self._SSO_CONFIG_VARS):
            return None

        config = {}
        missing_config_vars = []
        for config_var in self._SSO_CONFIG_VARS:
            if config_var in profile_config:
                config[config_var] = profile_config[config_var]
            else:
                missing_config_vars.append(config_var)

        if missing_config_vars:
            missing = ", ".join(missing_config_vars)
            raise InvalidConfigError(
                error_msg=(
                    f'The profile "{profile_name}" is configured to use SSO but is missing '
                    f"required configuration: {missing}"
                )
            )

        return config

    def load(self):
        """Load AWS SSO credentials."""
        sso_config = self._load_sso_config()
        if not sso_config:
            return None

        sso_fetcher = SSOCredentialFetcher(
            sso_config["sso_start_url"],
            sso_config["sso_region"],
            sso_config["sso_role_name"],
            sso_config["sso_account_id"],
            self._client_creator,
            token_loader=SSOTokenLoader(cache=self._token_cache),
            cache=self.cache,
        )

        return DeferredRefreshableCredentials(
            method=self.METHOD, refresh_using=sso_fetcher.fetch_credentials,
        )



"""Runway config test definition."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Generic, Tuple, TypeVar, Union

from ....variables import Variable
from ...models.runway import (
    CfnLintRunwayTestArgs,
    CfnLintRunwayTestDefinitionModel,
    RunwayTestDefinitionModel,
    ScriptRunwayTestArgs,
    ScriptRunwayTestDefinitionModel,
    ValidRunwayTestTypeValues,
    YamlLintRunwayTestDefinitionModel,
)
from .base import ConfigComponentDefinition

if TYPE_CHECKING:
    from typing_extensions import Literal

    from ...models.base import ConfigProperty


_DataModel = TypeVar(
    "_DataModel",
    CfnLintRunwayTestDefinitionModel,
    RunwayTestDefinitionModel,
    ScriptRunwayTestDefinitionModel,
    YamlLintRunwayTestDefinitionModel,
)


class RunwayTestDefinition(Generic[_DataModel], ConfigComponentDefinition):
    """Runway test definition."""

    args: Union[Dict[str, Any], ConfigProperty]
    name: str
    required: bool
    type: ValidRunwayTestTypeValues

    _data: RunwayTestDefinitionModel
    _supports_vars: Tuple[str, ...] = ("args", "required")

    def __init__(self, data: _DataModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    # error present on python3.7
    def __new__(  # pylint: disable=arguments-differ
        cls,
        data: _DataModel,
    ) -> RunwayTestDefinition[_DataModel]:
        """Create a new instance of a class.

        Returns:
            Correct subclass of RunwayTestDefinition for the given data.

        """
        if cls is not RunwayTestDefinition:
            return super().__new__(cls)
        if isinstance(data, CfnLintRunwayTestDefinitionModel):
            return super().__new__(CfnLintRunwayTestDefinition)
        if isinstance(data, ScriptRunwayTestDefinitionModel):
            return super().__new__(ScriptRunwayTestDefinition)
        if isinstance(data, YamlLintRunwayTestDefinitionModel):
            return super().__new__(YamlLintRunwayTestDefinition)
        raise TypeError(
            "expected data of type CfnLintRunwayTestDefinitionModel, "
            "ScriptRunwayTestDefinitionModel, or YamlLintRunwayTestDefinitionModel; "
            f"got {type(data)}"
        )

    def _register_variable(self, var_name: str, var_value: Any) -> None:
        """Register a variable.

        Args:
            var_name: Name of the config field that can contain a variable
                lookup.
            var_value: Literal value supplied in the config to be resolved
                as a variable if it contains a lookup.

        """
        self._vars[var_name] = Variable(
            name=f"{self.name}.{var_name}", value=var_value, variable_type="runway"
        )

    @classmethod
    def parse_obj(cls, obj: Any) -> RunwayTestDefinition[_DataModel]:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(RunwayTestDefinitionModel.parse_obj(obj))


class CfnLintRunwayTestDefinition(
    RunwayTestDefinition[CfnLintRunwayTestDefinitionModel]
):
    """Runway cfn-lint test definition."""

    args: CfnLintRunwayTestArgs
    type: Literal["cfn-lint"] = "cfn-lint"

    def __init__(self, data: CfnLintRunwayTestDefinitionModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    @classmethod
    def parse_obj(cls, obj: Any) -> CfnLintRunwayTestDefinition:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(CfnLintRunwayTestDefinitionModel.parse_obj(obj))


class ScriptRunwayTestDefinition(RunwayTestDefinition[ScriptRunwayTestDefinitionModel]):
    """Runway script test definition."""

    args: ScriptRunwayTestArgs
    type: Literal["script"] = "script"

    def __init__(self, data: ScriptRunwayTestDefinitionModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    @classmethod
    def parse_obj(cls, obj: Any) -> ScriptRunwayTestDefinition:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(ScriptRunwayTestDefinitionModel.parse_obj(obj))


class YamlLintRunwayTestDefinition(
    RunwayTestDefinition[YamlLintRunwayTestDefinitionModel]
):
    """Runway yamllint test definition."""

    type: Literal["yamllint"] = "yamllint"

    def __init__(self, data: YamlLintRunwayTestDefinitionModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    @classmethod
    def parse_obj(cls, obj: Any) -> YamlLintRunwayTestDefinition:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(YamlLintRunwayTestDefinitionModel.parse_obj(obj))



"""Delete sync strategy.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/syncstrategy/delete.py

"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar, Optional

from typing_extensions import Literal

from .base import BaseSync

if TYPE_CHECKING:
    from ..file_generator import FileStats


LOGGER = logging.getLogger(__name__.replace("._", "."))


class DeleteSync(BaseSync):
    """Delete file."""

    NAME: ClassVar[Literal["delete"]] = "delete"

    def determine_should_sync(
        self, src_file: Optional[FileStats], dest_file: Optional[FileStats]
    ) -> bool:
        """Determine if file should sync."""
        if dest_file:
            dest_file.operation_name = "delete"
        LOGGER.debug(
            "syncing: (None) -> %s (remove), file does not "
            "exist at source (%s) and delete mode enabled",
            dest_file.src if dest_file else None,
            dest_file.dest if dest_file else None,
        )
        return True



"""``runway init`` command."""
# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any, Tuple

import click
from pydantic import ValidationError

from ...core import Runway
from ...exceptions import ConfigNotFound, VariablesFileNotFound
from .. import options
from ..utils import select_deployments

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("init", short_help="initialize/bootstrap things")
@options.ci
@options.debug
@options.deploy_environment
@options.no_color
@options.tags
@options.verbose
@click.pass_context
def init(ctx: click.Context, debug: bool, tags: Tuple[str, ...], **_: Any) -> None:
    """Run initialization/bootstrap steps.

    \b
    Process
    -------
    1. Determines the deploy environment.
        - "-e, --deploy-environment" option
        - "DEPLOY_ENVIRONMENT" environment variable
        - git branch name
            - strips "ENV-" prefix, master is converted to common
            - ignored if "ignore_git_branch: true"
        - name of the current working directory
    2. Selects deployments & modules to deploy.
        - (default) prompts
        - (tags) module contains all tags
        - (non-interactive) all
    3. Initializes/bootstraps selected deployments/modules in the order defined.
       (e.g. "cdk bootstrap", "terraform init")

    \b
    Steps By Module Type
    --------------------
    - AWS CDK: Runs "cdk bootstrap".
    - CFNgin: Creates the "cfngin_bucket" if needed.
    - Terraform: Runs "terraform init", changes the workspace if needed, runs
      "terraform init" again if the workspace was changed, and finally
      downloads/updates Terraform modules.

    """  # noqa: D301
    try:
        Runway(ctx.obj.runway_config, ctx.obj.get_runway_context()).init(
            select_deployments(ctx, ctx.obj.runway_config.deployments, tags)
        )
    except ValidationError as err:
        LOGGER.error(err, exc_info=debug)
        ctx.exit(1)
    except (ConfigNotFound, VariablesFileNotFound) as err:
        LOGGER.error(err.message, exc_info=debug)
        ctx.exit(1)



"""Utility functions for S3."""
from __future__ import annotations

import logging
import os
import tempfile
import zipfile
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Sequence, cast

import boto3
from botocore.exceptions import ClientError

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.service_resource import S3ServiceResource
    from mypy_boto3_s3.type_defs import ObjectTypeDef

    from ._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


def _get_client(
    session: Optional[boto3.Session] = None, region: Optional[str] = None
) -> S3Client:
    """Get S3 boto client."""
    return session.client("s3") if session else boto3.client("s3", region_name=region)


def _get_resource(
    session: Optional[boto3.Session] = None, region: Optional[str] = None
) -> S3ServiceResource:
    """Get S3 boto resource."""
    return (
        session.resource("s3") if session else boto3.resource("s3", region_name=region)
    )


def purge_and_delete_bucket(
    bucket_name: str, region: str = "us-east-1", session: Optional[boto3.Session] = None
) -> None:
    """Delete all objects and versions in bucket, then delete bucket."""
    purge_bucket(bucket_name, region, session)
    delete_bucket(bucket_name, region, session)


def purge_bucket(
    bucket_name: str, region: str = "us-east-1", session: Optional[boto3.Session] = None
) -> None:
    """Delete all objects and versions in bucket."""
    if does_bucket_exist(bucket_name, region, session):
        s3_resource = _get_resource(session, region)
        bucket = s3_resource.Bucket(bucket_name)
        bucket.object_versions.delete()
    else:
        LOGGER.warning('bucket "%s" does not exist in region "%s"', bucket_name, region)


def delete_bucket(
    bucket_name: str, region: str = "us-east-1", session: Optional[boto3.Session] = None
) -> None:
    """Delete bucket."""
    if does_bucket_exist(bucket_name, region, session):
        LOGGER.verbose('delete bucket "%s"...', bucket_name)
        s3_resource = _get_resource(session, region)
        bucket = s3_resource.Bucket(bucket_name)
        bucket.delete()
        LOGGER.info('delete bucket "%s"', bucket_name)
    else:
        LOGGER.warning('bucket "%s" does not exist in region "%s"', bucket_name, region)


def does_bucket_exist(
    bucket_name: str, region: str = "us-east-1", session: Optional[boto3.Session] = None
) -> bool:
    """Check if bucket exists in S3."""
    s3_resource = _get_resource(session, region)
    try:
        s3_resource.meta.client.head_bucket(Bucket=bucket_name)
        return True
    except ClientError as exc:
        if exc.response["Error"]["Message"] == "Not Found":
            LOGGER.info('bucket "%s" does not exist', bucket_name)
            return False
        if exc.response["Error"]["Message"] == "Forbidden":
            LOGGER.exception(
                'access denied for bucket "%s" (permissions?)', bucket_name
            )
            raise
    return False


def ensure_bucket_exists(
    bucket_name: str, region: str = "us-east-1", session: Optional[boto3.Session] = None
) -> None:
    """Ensure S3 bucket exists."""
    if not does_bucket_exist(bucket_name, region, session):
        LOGGER.info('creating bucket "%s" (in progress)', bucket_name)
        s3_client = _get_client(session, region)
        if region == "us-east-1":
            create_bucket_opts: Dict[str, Any] = {}
        else:
            create_bucket_opts = {
                "CreateBucketConfiguration": {"LocationConstraint": region}
            }
        s3_client.create_bucket(Bucket=bucket_name, **create_bucket_opts)

        # sometimes when creating the bucket it can take a few moments before
        # it is ready to add the encryption settings.
        bucket_waiter = s3_client.get_waiter("bucket_exists")
        bucket_waiter.wait(Bucket=bucket_name)
        LOGGER.info('creating bucket "%s" (complete)', bucket_name)

        # enable default encryption
        s3_client.put_bucket_encryption(
            Bucket=bucket_name,
            ServerSideEncryptionConfiguration={
                "Rules": [
                    {"ApplyServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"}}
                ]
            },
        )
        LOGGER.verbose('enabled encryption for bucket "%s"', bucket_name)


def does_s3_object_exist(
    bucket: str,
    key: str,
    session: Optional[boto3.Session] = None,
    region: str = "us-east-1",
) -> bool:
    """Determine if object exists on s3."""
    s3_resource = _get_resource(session, region)
    try:
        s3_resource.Object(bucket, key).load()
        LOGGER.debug("s3 object exists: %s/%s", bucket, key)
    except ClientError as exc:
        if exc.response["Error"]["Code"] == "404":
            LOGGER.debug("s3 object does not exist: %s/%s", bucket, key)
            return False
        raise
    return True


def upload(
    bucket: str, key: str, filename: str, session: Optional[boto3.Session] = None
) -> None:
    """Upload file to S3 bucket."""
    s3_client = _get_client(session)
    LOGGER.info("uploading %s to s3://%s/%s...", filename, bucket, key)
    s3_client.upload_file(Filename=filename, Bucket=bucket, Key=key)


def download(
    bucket: str, key: str, file_path: str, session: Optional[boto3.Session] = None
) -> str:
    """Download a file from S3 to the given path."""
    s3_client = _get_client(session)

    LOGGER.info("downloading s3://%s/%s to %s...", bucket, key, file_path)
    s3_client.download_file(Bucket=bucket, Key=key, Filename=file_path)
    return file_path


def download_and_extract_to_mkdtemp(
    bucket: str, key: str, session: Optional[boto3.Session] = None
) -> str:
    """Download zip archive and extract it to temporary directory."""
    filedes, temp_file = tempfile.mkstemp()
    os.close(filedes)
    download(bucket, key, temp_file, session)

    output_dir = tempfile.mkdtemp()
    with zipfile.ZipFile(temp_file, "r") as zip_ref:
        zip_ref.extractall(output_dir)
    os.remove(temp_file)
    LOGGER.verbose("extracted %s to %s", temp_file, output_dir)
    return output_dir


def get_matching_s3_objects(
    bucket: str,
    prefix: Sequence[str] = "",
    suffix: str = "",
    session: Optional[boto3.Session] = None,
) -> Iterator[ObjectTypeDef]:
    """Generate objects in an S3 bucket.

    Args:
        bucket: Name of the S3 bucket.
        prefix: Only fetch objects whose key starts with
            this prefix (optional).
        suffix: Only fetch objects whose keys end with
            this suffix (optional).
        session: Boto3/botocore session.

    """
    s3_client = _get_client(session)
    paginator = s3_client.get_paginator("list_objects_v2")

    kwargs = {"Bucket": bucket}

    # We can pass the prefix directly to the S3 API.  If the user has passed
    # a tuple or list of prefixes, we go through them one by one.
    prefixes = (prefix,) if isinstance(prefix, str) else prefix
    for key_prefix in prefixes:
        kwargs["Prefix"] = key_prefix

        for page in paginator.paginate(**kwargs):
            try:
                contents = page["Contents"]
            except KeyError:
                return

            for obj in contents:
                if "Key" in obj and obj["Key"].endswith(suffix):
                    yield obj


def get_matching_s3_keys(
    bucket: str,
    prefix: str = "",
    suffix: str = "",
    session: Optional[boto3.Session] = None,
) -> Iterator[str]:
    """Generate the keys in an S3 bucket.

    Args:
        bucket: Name of the S3 bucket.
        prefix: Only fetch keys that start with this prefix (optional).
        suffix: Only fetch keys that end with this suffix (optional).
        session: Boto3/botocore session.

    """
    for obj in get_matching_s3_objects(bucket, prefix, suffix, session):
        if "Key" in obj:
            yield obj["Key"]



"""Constant values."""
AWS_SAM_BUILD_IMAGE_PREFIX = "public.ecr.aws/sam/build-"
"""Prefix for build image registries."""

DEFAULT_IMAGE_NAME = "runway.cfngin.hooks.awslambda"
"""Default name to apply to an image when building from a Dockerfile."""

DEFAULT_IMAGE_TAG = "latest"
"""Default tag to apply to an image when building from a Dockerfile."""



"""Base classes."""
from __future__ import annotations

import logging
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Literal

from ....compat import cached_property
from ..protocols import CfnginHookProtocol
from .exceptions import RuntimeMismatchError
from .models.args import AwsLambdaHookArgs
from .models.responses import AwsLambdaHookDeployResponse
from .source_code import SourceCode

if TYPE_CHECKING:
    from ...._logging import RunwayLogger
    from ....context import CfnginContext
    from ....utils import BaseModel
    from .deployment_package import DeploymentPackage
    from .docker import DockerDependencyInstaller
    from .type_defs import AwsLambdaHookDeployResponseTypedDict

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))

_AwsLambdaHookArgsTypeVar_co = TypeVar(
    "_AwsLambdaHookArgsTypeVar_co", bound=AwsLambdaHookArgs, covariant=True
)


class Project(Generic[_AwsLambdaHookArgsTypeVar_co]):
    """Project continaing source code for an AWS Lambda Function."""

    DEFAULT_CACHE_DIR_NAME: ClassVar[str] = "cache"
    """Name of the default cache directory."""

    args: _AwsLambdaHookArgsTypeVar_co
    """Parsed hook arguments."""

    ctx: CfnginContext
    """CFNgin context object."""

    def __init__(
        self, args: _AwsLambdaHookArgsTypeVar_co, context: CfnginContext
    ) -> None:
        """Instantiate class.

        Args:
            args: Parsed hook arguments.
            context: Context object.

        """
        self.args = args
        self.ctx = context

    @cached_property
    def build_directory(self) -> Path:
        """Directory being used to build deployment package."""
        result = (
            self.ctx.work_dir
            / f"{self.source_code.root_directory.name}.{self.source_code.md5_hash}"
        )
        result.mkdir(exist_ok=True, parents=True)
        return result

    @cached_property
    def cache_dir(self) -> Optional[Path]:
        """Directory where a dependency manager's cache data will be stored.

        Returns:
            Explicit cache directory if provided or default cache directory if
            it is not provided. If configured to not use cache, will always be
            ``None``.

        """
        if not self.args.use_cache:
            return None
        cache_dir = (
            self.args.cache_dir
            if self.args.cache_dir
            else self.ctx.work_dir / self.DEFAULT_CACHE_DIR_NAME
        )
        cache_dir.mkdir(exist_ok=True, parents=True)
        return cache_dir

    @cached_property
    def compatible_architectures(self) -> Optional[List[str]]:
        """List of compatible instruction set architectures."""
        return getattr(self.args, "compatible_architectures", None)

    @cached_property
    def compatible_runtimes(self) -> Optional[List[str]]:
        """List of compatible runtimes.

        Value should be valid Lambda Function runtimes
        (https://docs.aws.amazon.com/lambda/latest/dg/lambda-runtimes.html).

        Raises:
            ValueError: Defined or detected runtime is not in the list of
                compatible runtimes.

        """
        runtimes = getattr(self.args, "compatible_runtimes", cast(List[str], []))
        if runtimes and self.runtime not in runtimes:
            raise ValueError(
                f"runtime ({self.runtime}) not in compatible runtimes ({', '.join(runtimes)})"
            )
        return runtimes

    @cached_property
    def dependency_directory(self) -> Path:
        """Directory to use as the target of ``pip install --target``."""
        result = self.build_directory / "dependencies"
        result.mkdir(exist_ok=True, parents=True)
        return result

    @cached_property
    def license(self) -> Optional[str]:
        """Software license for the project.

        Can be any of the following:

        - A SPDX license identifier (e.g. ``MIT``).
        - The URL of a license hosted on the internet (e.g.
          ``https://opensource.org/licenses/MIT``).
        - The full text of the license.

        """
        return getattr(self.args, "license", None)

    @cached_property  # pylint error is python3.7 only
    def metadata_files(self) -> Tuple[Path, ...]:
        """Project metadata files (e.g. ``project.json``, ``pyproject.toml``)."""
        return ()

    @cached_property
    def runtime(self) -> str:
        """runtime of the build system.

        Value should be a valid Lambda Function runtime
        (https://docs.aws.amazon.com/lambda/latest/dg/lambda-runtimes.html).

        This property can be overwritten by subclasses when runtime can be
        determined through additional means.

        """
        if self._runtime_from_docker:
            return self._validate_runtime(self._runtime_from_docker)
        raise ValueError("runtime could not be determined from the build system")

    @cached_property
    def _runtime_from_docker(self) -> Optional[str]:
        """runtime from Docker if class can use Docker."""
        docker: Optional[DockerDependencyInstaller] = getattr(self, "docker", None)
        if not docker:
            return None
        return docker.runtime

    def _validate_runtime(self, detected_runtime: str) -> str:
        """Verify that the detected runtime matches what is explicitly defined.

        This method should be used before returning the detected runtime from
        the ``.runtime`` property.

        Args:
            detected_runtime: The runtime detected from the build system.

        Raises:
            RuntimeMismatchError: The detected runtime does not match what is
                defined.

        """
        if self.args.runtime and self.args.runtime != detected_runtime:
            raise RuntimeMismatchError(self.args.runtime, detected_runtime)
        return detected_runtime

    @cached_property
    def source_code(self) -> SourceCode:
        """Project source code.

        Lazy load source code object.
        Extends gitignore as needed.

        """
        source_code = SourceCode(
            self.args.source_code,
            include_files_in_hash=self.metadata_files,
            project_root=self.project_root,
        )
        for rule in self.args.extend_gitignore:
            source_code.add_filter_rule(rule)
        return source_code

    @cached_property
    def project_root(self) -> Path:
        """Root directory of the project.

        The top-level directory containing the source code and all
        configuration/metadata files (e.g. pyproject.toml, package.json).

        The project root can be different from the source code directory but,
        if they are different, the project root should contain the source code
        directory. If it does not, the source code directory will be always
        be used.

        The primary use case for this property is to allow configuration files
        to exist outside of the source code directory. The ``project_type``
        can and should rely on the value of this property when determining the
        type.

        """
        top_lvl_dir = (
            self.ctx.config_path.parent
            if self.ctx.config_path.is_file()
            else (
                self.ctx.config_path
                if self.ctx.config_path.is_dir()
                else self.args.source_code
            )
        )
        if top_lvl_dir == self.args.source_code:
            return top_lvl_dir

        parents = list(self.args.source_code.parents)
        if top_lvl_dir not in parents:
            LOGGER.info(
                "ignoring project directory; "
                "source code located outside of project directory"
            )
            return self.args.source_code

        dirs_to_check = [
            self.args.source_code,
            *parents[: parents.index(top_lvl_dir) + 1],
        ]
        for dir_to_check in dirs_to_check:
            for check_for_file in self.supported_metadata_files:
                if next(dir_to_check.glob(check_for_file), None):
                    return dir_to_check
        # reached if all dirs in between source and top-level are missing metadata files
        return top_lvl_dir

    @cached_property
    def project_type(self) -> str:
        """Type of project (e.g. poetry, yarn).

        This should be considered more of a "subtype" as the subclass should
        distinguish project language. The value of this property should reflect
        the project/dependency management tool used within the project.

        The value of this property should be calculated without initalizing
        other properties (e.g. ``source_code``) except for ``project_root``
        so that it can be used in their initialization process.

        """
        raise NotImplementedError

    @cached_property  # pylint error is python3.7 only
    def supported_metadata_files(self) -> Set[str]:
        """Names of all supported metadata files.

        Returns:
            Set of file names - not paths.

        """
        return set()

    def cleanup(self) -> None:
        """Cleanup project files at the end of execution.

        If any cleanup is needed (e.g. removal of temporary dependency directory)
        it should be implimented here. Hook's should call this method in a
        ``finally`` block to ensure it is run even if the rest of the hook
        encountered an error.

        """

    def cleanup_on_error(self) -> None:
        """Cleanup project files when an error occurs.

        This will be run before ``self.cleanup()`` if an error has occurred.

        Hooks should call this method in an ``except`` block and reraise the
        error afterward.

        """

    def install_dependencies(self) -> None:
        """Install project dependencies.

        Arguments/options should be read from the ``args`` attribute of this
        object instead of being passed into the method call. The method itself
        only exists for timing and filling in custom handling that is required
        for each project type.

        """
        raise NotImplementedError


_ProjectTypeVar = TypeVar("_ProjectTypeVar", bound=Project[AwsLambdaHookArgs])


class AwsLambdaHook(CfnginHookProtocol, Generic[_ProjectTypeVar]):
    """Base class for AWS Lambda hooks."""

    BUILD_LAYER: ClassVar[bool] = False
    """Flag to denote if the hook creates a Lambda Function or Layer deployment package."""

    args: AwsLambdaHookArgs
    """Parsed hook arguments."""

    ctx: CfnginContext
    """CFNgin context object."""

    # pylint: disable=super-init-not-called
    def __init__(self, context: CfnginContext, **_kwargs: Any) -> None:
        """Instantiate class.

        This method should be overridden by subclasses.
        This is required to set the value of the args attribute.

        Args:
            context: CFNgin context object (passed in by CFNgin).

        """
        self.ctx = context

    @cached_property
    def deployment_package(self) -> DeploymentPackage[_ProjectTypeVar]:
        """AWS Lambda deployment package."""
        raise NotImplementedError

    @cached_property
    def project(self) -> _ProjectTypeVar:
        """Project being deployed as an AWS Lambda Function."""
        raise NotImplementedError

    @overload
    def build_response(self, stage: Literal["deploy"]) -> AwsLambdaHookDeployResponse:
        ...

    @overload
    def build_response(self, stage: Literal["destroy"]) -> Optional[BaseModel]:
        ...

    @overload
    def build_response(self, stage: Literal["plan"]) -> AwsLambdaHookDeployResponse:
        ...

    def build_response(
        self, stage: Literal["deploy", "destroy", "plan"]
    ) -> Optional[BaseModel]:
        """Build response object that will be returned by this hook.

        Args:
            stage: The current stage being executed by the hook.

        """
        if stage == "deploy":
            return self._build_response_deploy()
        if stage == "destroy":
            return self._build_response_destroy()
        if stage == "plan":
            return self._build_response_plan()
        raise NotImplementedError("only deploy and destroy are supported")

    def _build_response_deploy(self) -> AwsLambdaHookDeployResponse:
        """Build response for deploy stage."""
        return AwsLambdaHookDeployResponse(
            bucket_name=self.deployment_package.bucket.name,
            code_sha256=self.deployment_package.code_sha256,
            compatible_architectures=self.deployment_package.compatible_architectures,
            compatible_runtimes=self.deployment_package.compatible_runtimes,
            license=self.deployment_package.license,
            object_key=self.deployment_package.object_key,
            object_version_id=self.deployment_package.object_version_id,
            runtime=self.deployment_package.runtime,
        )

    def _build_response_destroy(self) -> Optional[BaseModel]:
        """Build response for destroy stage."""
        return None

    def _build_response_plan(self) -> AwsLambdaHookDeployResponse:
        """Build response for plan stage."""
        try:
            return AwsLambdaHookDeployResponse(
                bucket_name=self.deployment_package.bucket.name,
                code_sha256=self.deployment_package.code_sha256,
                compatible_architectures=self.deployment_package.compatible_architectures,
                compatible_runtimes=self.deployment_package.compatible_runtimes,
                license=self.deployment_package.license,
                object_key=self.deployment_package.object_key,
                object_version_id=self.deployment_package.object_version_id,
                runtime=self.deployment_package.runtime,
            )
        except FileNotFoundError:
            return AwsLambdaHookDeployResponse(
                bucket_name=self.deployment_package.bucket.name,
                code_sha256="WILL CALCULATE WHEN BUILT",
                compatible_architectures=self.deployment_package.compatible_architectures,
                compatible_runtimes=self.deployment_package.compatible_runtimes,
                license=self.deployment_package.license,
                object_key=self.deployment_package.object_key,
                object_version_id=self.deployment_package.object_version_id,
                runtime=self.deployment_package.runtime,
            )

    def cleanup(self) -> None:
        """Cleanup temporary files at the end of execution.

        If any cleanup is needed (e.g. removal of temporary dependency directory)
        it should be implimented here. A Hook's stage methods should call this
        method in a ``finally`` block to ensure it is run even if the rest of
        the hook encountered an error.

        .. rubric:: Example
        .. code-block:: python

            def pre_deploy(self) -> Any:
                try:
                    pass  # primary logic
                except BaseException:
                    self.cleanup_on_error()
                    raise
                finally:
                    self.cleanup()

        """

    def cleanup_on_error(self) -> None:
        """Cleanup temporary files when an error occurs.

        This will be run before ``self.cleanup()`` if an error has occurred.

        A Hook's stage method should call this method in an ``except`` block
        and reraise the error afterward.

        .. rubric:: Example
        .. code-block:: python

            def pre_deploy(self) -> Any:
                try:
                    pass  # primary logic
                except BaseException:
                    self.cleanup_on_error()
                    raise
                finally:
                    self.cleanup()

        """

    def plan(self) -> AwsLambdaHookDeployResponseTypedDict:
        """Run during the **plan** stage."""
        return cast(
            "AwsLambdaHookDeployResponseTypedDict",
            self.build_response("plan").dict(by_alias=True),
        )

    def post_deploy(self) -> Any:
        """Run during the **post_deploy** stage."""
        LOGGER.warning("post_deploy not implimented for %s", self.__class__.__name__)
        return True

    def post_destroy(self) -> Any:
        """Run during the **post_destroy** stage."""
        LOGGER.warning("post_destroy not implimented for %s", self.__class__.__name__)
        return True

    def pre_deploy(self) -> Any:
        """Run during the **pre_deploy** stage."""
        LOGGER.warning("pre_deploy not implimented for %s", self.__class__.__name__)
        return True

    def pre_destroy(self) -> Any:
        """Run during the **pre_destroy** stage."""
        LOGGER.warning("pre_destroy not implimented for %s", self.__class__.__name__)
        return True



"""Version utilities."""
from __future__ import annotations

import packaging.version


class Version(packaging.version.Version):
    """Customize packagining.version.Version."""

    def __init__(self, version: str) -> None:
        """Instantiate class.

        Args:
            version: Version string. (e.g. 1.0.0, v1.0.0)

        """
        self._original_text = version
        super().__init__(version)

    def __repr__(self) -> str:
        """Return repr."""
        # this usage of super is required to reproduce the intended result in
        # any subclasses of this class
        # pylint: disable=super-with-arguments
        return f"<Version('{super(Version, self).__str__()}')>"

    def __str__(self) -> str:
        """Return the original version string."""
        return self._original_text
